#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 13 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ....viewer.modules.roi_manager_modules.roi_list import ROIList
from .roi_types import *
from ...core.common import ViewerUtils
from .... import pyqtgraphCore as pg
from copy import deepcopy
from .read_imagej import read_roi_zip as read_imagej
from ....common.configuration import HAS_CAIMAN
from matplotlib import cm as matplotlib_color_map
from tqdm import tqdm
import logging
from itertools import product as iter_product

if HAS_CAIMAN:
    from caiman.utils.visualization import get_contours as caiman_get_contours
    from caiman.source_extraction.cnmf.cnmf import load_CNMF

logger = logging.getLogger()


class AbstractBaseManager(metaclass=abc.ABCMeta):
    """Base ROI Manager"""
    def __init__(self, parent, ui, viewer_interface: ViewerUtils):
        """
        Set the common attributes

        :param parent: The ModuleGUI QDockWidget instance
        :param ui:  The ui of the ModuleGUI QDockWidget instance,
        :param viewer_interface:    A ViewerUtils instance for accessing the Viewer the parent QDockWidget belongs to
        """
        self.ui = ui
        self.vi = viewer_interface
        self.roi_list = None  #: The ROIList instance that stores the list of ROIs
        self.parent = parent

        self.metadata = None

    @abc.abstractmethod
    def add_roi(self, *args, **kwargs):
        """Method for adding an ROI, must be implemented in subclass"""
        pass

    def is_empty(self) -> bool:
        """Return true if the ROI list is empty, else return False"""
        if not hasattr(self, 'roi_list'):
            return True
        if self.roi_list is None:
            return True
        if len(self.roi_list) < 1:
            return True
        else:
            return False

    def get_all_states(self) -> dict:
        """
        Get the ROI states for all ROIs in self.roi_list so that they can be restored.
        The appropriate manager is instantiated based on the 'roi_type' key of the returned dict
        """
        self.vi.viewer.status_bar_label.showMessage('Saving ROIs...')
        # the key 'roi_type' determines which Manager subclass should be used, and 'states' are the actual ROI states
        states = {
            'roi_type': self.roi_list.roi_types.__name__,
            'states': [],
            'metadata': self.metadata
        }

        for roi in self.roi_list:
            state = roi.to_state()
            states['states'].append(state)
        self.vi.viewer.status_bar_label.showMessage('ROIs saved!')
        return states

    def restore_from_states(self, states: dict):
        if 'metadata' in states.keys():
            self.metadata = states['metadata']

    def get_plot_item(self) -> pg.PlotDataItem:
        """Get the viewer plot item that is associated to these ROIs"""
        return self.vi.viewer.ui.roiPlot.plot()

    def clear(self):
        """Cleanup of all ROIs in the list"""
        if not hasattr(self, 'roi_list'):
            return
        self.roi_list.clear_()
        del self.roi_list

    def __del__(self):
        """Cleanup of all ROIs in the list and deletes the manager instance. Used when switching modes."""
        self.clear()
        # self.roi_list.list_widget.clear()
        # self.roi_list.list_widget_tags.clear()
        # self.roi_list.disconnect_all()
        # for i in range(len(self.roi_list)):


class ManagerManual(AbstractBaseManager):
    """The Manager for the Manual mode"""
    def __init__(self, parent, ui, viewer_interface):
        super(ManagerManual, self).__init__(parent, ui, viewer_interface)
        self.create_roi_list()

    def create_roi_list(self):
        """Create a new empty ROI list instance for storing Manual ROIs"""
        self.roi_list = ROIList(self.ui, ManualROI, self.vi)

    def add_roi(self, shape: str) -> ManualROI:
        """
        Add an ROI to the list

        :param shape: either "PolyLineROI" or "EllipseROI"
        """
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        dims = self.vi.viewer.workEnv.imgdata.seq.shape
        roi_graphics_object = ManualROI.get_generic_roi_graphics_object(shape, dims)

        roi = ManualROI(self.get_plot_item(), roi_graphics_object, self.vi.viewer.getView())

        self.roi_list.append(roi)
        self.roi_list.reindex_colormap()

        self.vi.workEnv_changed("ROIs imported")

        return roi

    def restore_from_states(self, states: dict):
        """Restore ROIs from states"""
        super(ManagerManual, self).restore_from_states(states)
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        for state in states['states']:
            roi = ManualROI.from_state(self.get_plot_item(), self.vi.viewer.getView(), state)
            self.roi_list.append(roi)
        # ix = 0
        # for state in states['states']:
        #     self.roi_list[ix].set_roi_graphics_object_state(state['roi_graphics_object_state'])
        #     ix += 1
        self.roi_list.reindex_colormap()

    def get_all_states(self) -> dict:
        """Get the ROI states so that they can be restored later"""
        self.vi.viewer.status_bar_label.showMessage('Saving ROIs...')

        states = {
            'roi_type': self.roi_list.roi_types.__name__,
            'states': [],
            'metadata': self.metadata
        }

        for ix in range(len(self.roi_list)):
            self.roi_list.set_pg_roi_plot(ix)
            state = self.roi_list[ix].to_state()
            states['states'].append(state)
        self.vi.viewer.status_bar_label.showMessage('Finished saving ROIs!')
        return states

    def import_from_imagej(self, path: str):
        """
        Uses read-roi package created by Hadrien Mary.
        https://pypi.org/project/read-roi/

        :param path: Full path to the ImageJ ROIs zip file
        """
        ij_roi = read_imagej(path)
        for k in ij_roi.keys():
            if ij_roi[k]['type'] in ('oval', 'rectangle'):
                width=ij_roi[k]['width']
                height=ij_roi[k]['height']
                left=ij_roi[k]['left']
                top=ij_roi[k]['top']
                if ij_roi[k]['type'] == 'oval':
                    x_bottom=left
                    y_bottom=top+height
                    ps=[x_bottom,y_bottom,width,height]
                    roi= ManualROI.from_ellipse(positions=ps,
                                           curve_plot_item=self.get_plot_item(),
                                           view_box=self.vi.viewer.getView())
                else:
                    ps=list(iter_product((left,left+width),(top,top+height)))
                    ps[2],ps[3]=ps[3],ps[2] #swap coordinates so it draws coordinates in sequence 
            elif (ij_roi[k]['type']=='freehand'):
                #freehand ROIs have large number of datapoints, so reducing it by a fifth
                xs = ij_roi[k]['x'][::5]
                ys = ij_roi[k]['y'][::5]
                ps = list(zip(xs, ys))
            else:
                xs = ij_roi[k]['x']
                ys = ij_roi[k]['y']
                ps = list(zip(xs, ys))                                      
            if ij_roi[k]['type'] != 'oval':
                roi = ManualROI.from_positions(positions=ps,
                                           curve_plot_item=self.get_plot_item(),
                                           view_box=self.vi.viewer.getView())
            self.roi_list.append(roi)
        self.roi_list.reindex_colormap()

    def add_roi_from_points(self, xs, ys):
        ps = list(zip(xs, ys))

        roi = ManualROI.from_positions(
            positions=ps,
            curve_plot_item=self.get_plot_item(),
            view_box=self.vi.viewer.getView()
        )

        self.roi_list.append(roi)
        self.roi_list.reindex_colormap()


class ManagerScatterROI(AbstractBaseManager):
    """Manager for unmoveable ROIs drawn using scatterplots"""
    def __init__(self, parent, ui, viewer_interface: ViewerUtils):
        super(ManagerScatterROI, self).__init__(parent, ui, viewer_interface)
        self.create_roi_list()
        self.list_widget = self.roi_list.list_widget

    def add_roi(self, curve: np.ndarray,
                xs: np.ndarray,
                ys: np.ndarray,
                metadata: dict = None,
                dfof_data: np.ndarray = None,
                spike_data: np.ndarray = None) \
            -> ScatterROI:
        """
        Add a single ROI

        `xs` and `ys` arguments are 1D numpy arrays.

        :param curve:       curve data, 1-D array, y values/intensity values
        :param xs:          x-values for the scatter plot to spatially illustrate the ROI
        :param ys:          corresponding y-values for the scatter plot to spatially illustrate the ROI
        :param metadata:    Any metadata for this ROI
        :return:            ScatterROI object
        """
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        roi = ScatterROI(
            curve_plot_item=self.get_plot_item(),
            view_box=self.vi.viewer.getView(),
            curve_data=curve,
            xs=xs,
            ys=ys,
            dfof_data=dfof_data,
            spike_data=spike_data
        )

        roi.metadata = metadata
        self.roi_list.append(roi)
        self.roi_list.reindex_colormap()

        self.vi.workEnv_changed("ROI Added")

        return roi

    def restore_from_states(self, states: dict):
        """Restore from states, such as when these ROIs are saved with a Project Sample"""
        super(ManagerScatterROI, self).restore_from_states(states)

        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        for state in states['states']:
            roi = ScatterROI.from_state(self.get_plot_item(), self.vi.viewer.getView(), state)
            self.roi_list.append(roi)

        self.roi_list.reindex_colormap()

    def create_roi_list(self):
        """Create empty ROI List"""
        self.roi_list = ROIList(self.ui, ScatterROI, self.vi)

    def set_spot_size(self, size: int):
        """Set the spot size for the scatter plot which illustrates the ROI"""
        for roi in self.roi_list:
            roi.get_roi_graphics_object().setSize(size)
            roi.spot_size = size


class ManagerVolROI(ManagerScatterROI):
    """Manager for 3D ROIs"""
    def __init__(self, parent, ui, viewer_interface: ViewerUtils):
        super(ManagerVolROI, self).__init__(parent, ui, viewer_interface)
        self.vi.viewer.sigZLevelChanged.connect(self.set_zlevel)

    def set_zlevel(self, z: int):
        """Set the current z-level to be visible in the viewer"""
        # if not hasattr(self, 'roi_list'):
        #     warn('roi list does not exist, probably empty work environment')
        #     return

        for roi in self.roi_list:
            roi.set_zlevel(z)

    def create_roi_list(self):
        """Create new empty ROI list"""
        self.roi_list = ROIList(self.ui, VolCNMF, self.vi)


class ManagerVolCNMF(ManagerVolROI):
    """Manager for 3D CNMF based ROIs"""
    def __init__(self, parent, ui, viewer_interface):
        super(ManagerVolCNMF, self).__init__(parent, ui, viewer_interface)
        self.roi_list = ROIList(self.ui, VolCNMF, self.vi)

        self.input_params_dict = None
        self.idx_components = None  # Keep track of components if the user manually want to remove some
        self.orig_idx_components = None  # List of components prior to any manual deletion by the user

        # cnmf data dict directly from the hdf5 file
        self.cnmf_data_dict = None

        # These correspond to the caiman.source_extraction.cnmf attributes
        self.cnmA = None
        self.cnmb = None
        self.cnmC = None
        self.cnm_f = None
        self.cnmYrA = None

    def create_roi_list(self):
        self.roi_list = ROIList(self.ui, VolCNMF, self.vi)

    def add_all_components(self, cnmf_data_dict: dict, input_params_dict: dict):
        """
        Add all components from a CNMF(E) output. Arguments correspond to CNMF(E) outputs

        :param cnmf_data_dict:      CNMF results data directly from the HDF5 file
        :param input_params_dict:   dict of input params, from the batch manager
        :param calc_raw_min_max:    Calculate raw min & max for each ROI
        :return:
        """
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        self.cnmf_data_dict = cnmf_data_dict

        # self.cnmf_obj = load_CNMF(self.cnmf_data_dict)

        self.cnmA = self.cnmf_data_dict['estimates']['A']
        self.cnmb = self.cnmf_data_dict['estimates']['b']
        self.cnm_f = self.cnmf_data_dict['estimates']['f']
        self.cnmC = self.cnmf_data_dict['estimates']['C']
        self.cnmYrA = self.cnmf_data_dict['estimates']['YrA']
        self.dims = self.cnmf_data_dict['dims']
        self.cnmS = self.cnmf_data_dict['estimates']['S']
        self.cnm_dfof = self.cnmf_data_dict['estimates']['F_dff']

        # components are already filtered from the output file
        self.idx_components = np.arange(self.cnmC.shape[0])
        self.orig_idx_components = deepcopy(self.idx_components)
        self.input_params_dict = input_params_dict

        # spatial components
        contours = caiman_get_contours(self.cnmA[:, self.idx_components], self.dims, thr=0.9)

        temporal_components = self.cnmC

        self.input_params_dict = self.input_params_dict
        num_components = len(temporal_components)

        self.ui.radioButton_curve_data.setChecked(True)

        for ix in range(num_components):
            self.vi.viewer.status_bar_label.showMessage('Please wait, adding component #: '
                                                        + str(ix) + ' / ' + str(num_components))

            curve_data = temporal_components[self.idx_components[ix]]
            contour = contours[ix]

            roi = VolCNMF(curve_plot_item=self.get_plot_item(),
                          view_box=self.vi.viewer.getView(),
                          cnmf_idx=self.idx_components[ix],
                          curve_data=curve_data,
                          contour=contour,
                          dfof_data=self.cnm_dfof[ix] if (self.cnm_dfof is not None) else None,
                          spike_data=self.cnmS[ix])

            self.roi_list.append(roi)

        self.vi.workEnv_changed("ROIs imported")
        self.roi_list.reindex_colormap(random_shuffle=True)
        self.vi.viewer.status_bar_label.showMessage('Finished adding all components!')

    def add_roi(self):
        """Not implemented, uses add_all_components to import all ROIs instead"""
        raise NotImplementedError('Not implemented for CNMFE ROIs')

    def restore_from_states(self, states: dict):
        """Restore from states, such as when these ROIs are saved with a Project Sample"""
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        self.cnmf_data_dict = states['cnmf_data_dict']

        for state in states['states']:
            roi = VolCNMF.from_state(self.get_plot_item(), self.vi.viewer.getView(), state)
            self.roi_list.append(roi)
        self.input_params_dict = states['input_params_cnmfe']
        self.cnmA = states['cnmf_output']['cnmA']
        self.cnmb = states['cnmf_output']['cnmb']
        self.cnmC = states['cnmf_output']['cnmC']
        self.cnm_f = states['cnmf_output']['cnm_f']
        self.cnmS = self.cnmf_data_dict['estimates']['S']
        self.cnm_dfof = self.cnmf_data_dict['estimates']['F_dff']
        self.cnmYrA = states['cnmf_output']['cnmYrA']
        self.idx_components = states['cnmf_output']['idx_components']
        self.orig_idx_components = states['cnmf_output']['orig_idx_components']
        self.roi_list.reindex_colormap()

    def get_all_states(self) -> dict:
        """Get all states so that they can be restored"""
        states = super(ManagerVolROI, self).get_all_states()

        # If the user has manually deleted some ROIs
        new_idx_components = np.array([roi.cnmf_idx for roi in self.roi_list], dtype=np.int64)

        # Make sure nothing weird happened
        # l = [self.cnmA, self.cnmb, self.cnmC, self.cnm_f, self.cnmYrA, self.orig_idx_components, new_idx_components]
        # if any(item is None for item in l):
        #     raise ValueError('One or more pieces of CNMF(E) data are missing')

        # Store the actual cnmf attributes as well.
        input_dict = {'input_params_cnmfe': self.input_params_dict,
                      'cnmf_data_dict':     self.cnmf_data_dict,
                      'cnmf_output':
                          {
                              'cnmA': self.cnmA,
                              'cnmb': self.cnmb,
                              'cnmC': self.cnmC,
                              'cnm_f': self.cnm_f,
                              'cnmYrA': self.cnmYrA,
                              'idx_components': new_idx_components,
                              'orig_idx_components': self.orig_idx_components
                          }
                      }

        states.update(input_dict)
        return states

    def update_idx_components(self, ix: int):
        """Update idx_components if the user manually delete an ROI"""
        roi = self.roi_list[self.roi_list.current_index]
        self.idx_components = np.delete(self.idx_components, np.where(self.idx_components == roi.cnmf_idx)[0])

    def set_spot_size(self, size: int):
        for roi in self.roi_list:
            roi.get_roi_graphics_object().setSize(size)
            roi.spot_size = size


class ManagerVolMultiCNMFROI(ManagerVolROI):
    """Manager for 3D data using one CNMF per plane"""
    def __init__(self, parent, ui, viewer_interface):
        super(ManagerVolMultiCNMFROI, self).__init__(parent, ui, viewer_interface)

        self.roi_list = ROIList(self.ui, VolMultiCNMFROI, self.vi)

        self.input_params_dict: dict = None
        self.idx_components: List[np.ndarray] = []
        self.orig_idx_components: List[np.ndarray] = []

        self.cnmf_data_dicts: List[dict] = []

        self.cnmA: List[np.ndarray] = []
        self.cnmb: List[np.ndarray] = []
        self.cnmC: List[np.ndarray] = []
        self.cnm_f: List[np.ndarray] = []
        self.cnmYrA: List[np.ndarray] = []
        self.cnmS: List[np.ndarray] = []
        self.cnm_dfof: List[np.ndarray] = []
        self.dims: List[tuple] = []

        self.roi_xys: List[np.ndarray] = []  # roi x-y coordinates
        self.roi_ixs: List[np.ndarray] = []  # the roi index that each coordinate maps to
        self.roi_crs: List[np.ndarray] = []  # the color that each roi index maps to
        self.roi_sps: List[pg.ScatterPlotItem] = []  # ROIs represented as scatterplots

        self.num_zlevels: int = 0

    def create_roi_list(self):
        self.roi_list = ROIList(self.ui, VolMultiCNMFROI, self.vi)

    def add_all_components(
            self,
            cnmf_data_dicts: List[dict],
            input_params_dict: dict,
    ):
        self.input_params_dict = input_params_dict

        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        self.cnmf_data_dicts = cnmf_data_dicts

        self.num_zlevels = len(self.cnmf_data_dicts)

        for zcenter, cnmf_data_dict in enumerate(self.cnmf_data_dicts):
            self.cnmA.append(cnmf_data_dict['estimates']['A'])
            self.cnmb.append(cnmf_data_dict['estimates']['b'])
            self.cnm_f.append(cnmf_data_dict['estimates']['f'])
            self.cnmC.append(cnmf_data_dict['estimates']['C'])
            self.cnmS.append(cnmf_data_dict['estimates']['S'])
            self.cnm_dfof.append(cnmf_data_dict['estimates']['F_dff'])
            self.cnmYrA.append(cnmf_data_dict['estimates']['YrA'])
            self.dims.append(cnmf_data_dict['dims'])
            self.idx_components.append(cnmf_data_dict['estimates']['idx_components'])

            if self.idx_components[-1] is None:
                self.idx_components[-1] = np.arange(self.cnmC[-1].shape[0])

            self.orig_idx_components.append(
                deepcopy(
                    self.idx_components[-1]
                )
            )

            contours = caiman_get_contours(
                self.cnmA[-1][:, self.idx_components[-1]],
                self.dims[-1],
                # swap_dim=True
            )

            num_components = len(self.idx_components[-1])

            self.ui.radioButton_curve_data.setChecked(True)

            roi_ixs = []
            roi_xy = []

            for ix in range(len(contours)):
                coors = contours[ix]['coordinates']
                coors = coors[~np.isnan(coors).any(axis=1)]
                roi_xy += [coors]
                roi_ixs += [ix] * coors.shape[0]

            roi_xy = np.vstack(roi_xy)
            roi_ixs = np.vstack(roi_ixs)

            self.roi_xys.append(roi_xy)
            self.roi_ixs.append(roi_ixs)

            cm = matplotlib_color_map.get_cmap('hsv')
            cm._init()
            lut = (cm._lut * 255).view(np.ndarray)

            cm_ixs = np.linspace(0, 210, np.unique(roi_ixs).size + 1, dtype=int)

            roi_crs = []

            for roi_ix, cm_ix in zip(np.unique(roi_ixs), cm_ixs):
                c = lut[cm_ix]
                roi_crs.append(
                    np.array([c] * roi_ixs[roi_ixs == roi_ix].size)  # color for each spot
                )

            roi_crs = np.vstack(roi_crs)
            self.roi_crs.append(roi_crs)

            xy_coors = self.roi_xys[-1]

            brushes = list(map(pg.mkBrush, roi_crs))
            pens = list(map(pg.mkPen, roi_crs))

            sp = pg.ScatterPlotItem(
                xy_coors[:, 0],
                xy_coors[:, 1],
                symbol='s',
                size=1,
                pxMode=True,
                brush=brushes,
                pen=pens
            )

            self.vi.viewer.getView().addItem(sp)
            sp.hide()
            self.roi_sps.append(sp)

            for ix in range(num_components):
                self.vi.viewer.status_bar_label.showMessage(
                    f"Please wait, adding component {ix} / {num_components} "
                    f"on zlevel {zcenter} / {self.num_zlevels - 1}"
                )

                curve_data = self.cnmC[-1][self.idx_components[-1][ix]]
                contour = contours[ix]

                if self.cnm_dfof[-1] is not None:
                    dfof_data = self.cnm_dfof[-1][ix]
                else:
                    dfof_data = None

                roi = VolMultiCNMFROI(
                    curve_plot_item=self.get_plot_item(),
                    view_box=self.vi.viewer.getView(),
                    cnmf_idx=self.idx_components[-1][ix],
                    curve_data=curve_data,
                    contour=contour,
                    spike_data=self.cnmS[-1][ix],
                    dfof_data=dfof_data,
                    zcenter=zcenter,
                    zlevel=self.vi.viewer.current_zlevel,
                    roi_ix=ix,
                    scatter_plot=sp,
                    parent_manager=self,
                )

                self.roi_list.append(roi, add_to_list_widget=False)

        self.roi_list.list_widget.addItems(
            list(map(str, range(len(self.roi_list))))
        )

        self.vi.workEnv_changed("ROIs imported")

        # self.roi_list.reindex_colormap(random_shuffle=True)
        self.roi_sps[self.vi.viewer.current_zlevel].show()

        self.vi.viewer.status_bar_label.showMessage('Finished adding all components!')

    def set_zlevel(self, z: int):
        """Set the current z-level to be visible in the viewer"""
        # super(ManagerVolMultiCNMFROI, self).set_zlevel(z)
        for roi in self.roi_list:
            roi.set_zlevel(z)

        for i in range(len(self.roi_sps)):
            if i == z:
                self.roi_sps[i].show()
            else:
                self.roi_sps[i].hide()

    def add_roi(self):
        """Not implemented, uses add_all_components to import all ROIs instead"""
        raise NotImplementedError('Not implemented for CNMFE ROIs')
    
    def restore_from_states(self, states: dict):
        if 'metadata' in states.keys():
            self.metadata = states['metadata']

        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        self.cnmf_data_dicts = states['cnmf_data_dicts']

        # for state in states['states']:
        #     roi = VolMultiCNMFROI.from_state(
        #         self.get_plot_item(),
        #         self.vi.viewer.getView(),
        #         state
        #     )
        #
        #     self.roi_list.append(roi)

        self.input_params_dict = states['input_params_cnmf']
        self.num_zlevels = states['num_zlevels']

        self.cnmA = states['cnmf_output']['cnmA']
        self.cnmb = states['cnmf_output']['cnmb']
        self.cnmC = states['cnmf_output']['cnmC']
        self.cnm_f = states['cnmf_output']['cnm_f']
        self.cnmYrA = states['cnmf_output']['cnmYrA']
        self.idx_components = states['cnmf_output']['idx_components']
        self.orig_idx_components = states['cnmf_output']['orig_idx_components']
        self.dims = states['cnmf_output']['dims']

        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        for zcenter in range(self.num_zlevels):
            logger.info(f"Loading z-level {zcenter}")
            contours = caiman_get_contours(
                self.cnmA[zcenter][:, self.idx_components[zcenter]],
                self.dims[zcenter],
                # swap_dim=True
            )

            num_components = len(self.idx_components[zcenter])

            self.ui.radioButton_curve_data.setChecked(True)

            roi_ixs = []
            roi_xy = []

            for ix in range(len(contours)):
                coors = contours[ix]['coordinates']
                coors = coors[~np.isnan(coors).any(axis=1)]
                roi_xy += [coors]
                roi_ixs += [ix] * coors.shape[0]

            roi_xy = np.vstack(roi_xy)
            roi_ixs = np.vstack(roi_ixs)

            self.roi_xys.append(roi_xy)
            self.roi_ixs.append(roi_ixs)

            cm = matplotlib_color_map.get_cmap('hsv')
            cm._init()
            lut = (cm._lut * 255).view(np.ndarray)

            cm_ixs = np.linspace(0, 210, np.unique(roi_ixs).size + 1, dtype=int)

            roi_crs = []

            for roi_ix, cm_ix in zip(np.unique(roi_ixs), cm_ixs):
                c = lut[cm_ix]
                roi_crs.append(
                    np.array([c] * roi_ixs[roi_ixs == roi_ix].size)  # color for each spot
                )

            roi_crs = np.vstack(roi_crs)
            self.roi_crs.append(roi_crs)

            xy_coors = self.roi_xys[-1]

            brushes = list(map(pg.mkBrush, roi_crs))
            pens = list(map(pg.mkPen, roi_crs))

            sp = pg.ScatterPlotItem(
                xy_coors[:, 0],
                xy_coors[:, 1],
                symbol='s',
                size=1,
                pxMode=True,
                brush=brushes,
                pen=pens
            )

            self.vi.viewer.getView().addItem(sp)
            sp.hide()
            self.roi_sps.append(sp)

            for ix in tqdm(range(len(self.idx_components[zcenter]))):
                self.vi.viewer.status_bar_label.showMessage(
                    f"Please wait, adding component {ix} / {num_components} "
                    f"on zlevel {zcenter} / {self.num_zlevels - 1}"
                )

                curve_data = self.cnmC[zcenter][self.idx_components[zcenter][ix]]
                contour = contours[ix]

                cnmf_idx = self.idx_components[zcenter][ix]

                roi = VolMultiCNMFROI(
                    curve_plot_item=self.get_plot_item(),
                    view_box=self.vi.viewer.getView(),
                    cnmf_idx=cnmf_idx,
                    curve_data=curve_data,
                    contour=contour,
                    zcenter=zcenter,
                    zlevel=self.vi.viewer.current_zlevel,
                    roi_ix=ix,
                    scatter_plot=sp,
                    parent_manager=self,
                )

                roi_state = list(
                    filter(
                        lambda r: r['cnmf_idx'] == cnmf_idx,
                        states['states'][zcenter]
                    )
                )[0]

                for k in roi_state['tags'].keys():
                    roi.set_tag(k, roi_state['tags'][k])

                roi.dfof_data = roi_state['dfof_data']
                roi.spike_data = roi_state['spike_data']

                self.roi_list.append(roi)

        self.roi_list.list_widget.addItems(
            list(map(str, range(len(self.roi_list))))
        )

        self.vi.workEnv_changed("ROIs imported")

        # self.roi_list.reindex_colormap(random_shuffle=True)
        self.roi_sps[self.vi.viewer.current_zlevel].show()

        self.vi.viewer.status_bar_label.showMessage('Finished adding all components!')

    def get_all_states(self) -> dict:
        roi_list_sorted = \
        [
            [
                roi for roi in self.roi_list if roi.zcenter == zlevel
            ] for zlevel in range(self.num_zlevels)
        ]

        states = \
            {
                'roi_type': self.roi_list.roi_types.__name__,
                'states': [],
                'metadata': self.metadata
            }

        roi_states = \
        [
            [
                roi.to_state() for roi in rois_zlevel
            ] for rois_zlevel in roi_list_sorted
        ]

        states['states'] = roi_states

        # make a new idx_components list in case the user has manually deleted some ROIs
        new_idx_components: List[np.ndarray] = []
        for zlevel in range(len(roi_list_sorted)):
            roi_cnmf_idxs = [roi.cnmf_idx for roi in roi_list_sorted[zlevel]]
            roi_cnmf_idxs.sort()
            new_idx_components.append(
                np.array(roi_cnmf_idxs, dtype=np.uint64)
            )

        # store the cnmf attributes as well
        input_dict = \
            {
                'input_params_cnmf': self.input_params_dict,
                'cnmf_data_dicts': self.cnmf_data_dicts,
                'num_zlevels': self.num_zlevels,
                'cnmf_output':
                    {
                        'cnmA': self.cnmA,
                        'cnmb': self.cnmb,
                        'cnmC': self.cnmC,
                        'cnm_f': self.cnm_f,
                        'cnmYrA': self.cnmYrA,
                        'idx_components': new_idx_components,
                        'orig_idx_components': self.orig_idx_components,
                        'dims': self.dims
                    }
            }

        states.update(input_dict)

        return states


class ManagerCNMFROI(AbstractBaseManager):
    """Manager for ROIs imported from CNMF or CNMFE outputs"""
    def __init__(self, parent, ui, viewer_interface):
        """Instantiate necessary attributes"""
        super(ManagerCNMFROI, self).__init__(parent, ui, viewer_interface)

        self.create_roi_list()
        self.list_widget = self.roi_list.list_widget
        self.input_params_dict = None
        self.idx_components = None  # Keep track of components if the user manually want to remove some
        self.orig_idx_components = None  # List of components prior to any manual deletion by the user

        # cnmf data dict directly from the hdf5 file
        self.cnmf_data_dict = None

        # These correspond to the caiman.source_extraction.cnmf attributes
        self.cnmA = None
        self.cnmb = None
        self.cnmC = None
        self.cnm_f = None
        self.cnmYrA = None

        self.raw_normalization_choices = ['top_5', 'top_10', 'top_5p', 'top_10p', 'top_25p']

    def create_roi_list(self):
        """Create empty CNMFROI list"""
        self.roi_list = ROIList(self.ui, CNMFROI, self.vi)

    def add_all_components(self, cnmf_data_dict, input_params_dict, calc_raw_min_max=False):
        """
        Add all components from a CNMF(E) output. Arguments correspond to CNMF(E) outputs

        :param cnmf_data_dict:      CNMF results data directly from the HDF5 file
        :param input_params_dict:   dict of input params, from the batch manager
        :param calc_raw_min_max:    Calculate raw min & max for each ROI
        :return:
        """
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        self.cnmf_data_dict = cnmf_data_dict

        self.cnmA = self.cnmf_data_dict['estimates']['A']
        self.cnmb = self.cnmf_data_dict['estimates']['b']
        self.cnm_f = self.cnmf_data_dict['estimates']['f']
        self.cnmC = self.cnmf_data_dict['estimates']['C']
        self.cnmS = self.cnmf_data_dict['estimates']['S']

        # # h5py is doing a weird thing where ``None`` gets stored as a byte string
        # if isinstance(self.cnmf_data_dict['estimates']['F_dff'][()], np.ndarray):
        #     self.cnm_dfof = self.cnmf_data_dict['estimates']['F_dff']
        # else:
        #     self.cnm_dfof = None


        self.cnm_dfof = self.cnmf_data_dict['estimates']['F_dff']
        self.cnmYrA = self.cnmf_data_dict['estimates']['YrA']
        self.dims = self.cnmf_data_dict['dims']
        self.idx_components = cnmf_data_dict['estimates']['idx_components']

        if self.idx_components is None:
            self.idx_components = np.arange(self.cnmC.shape[0])

        self.orig_idx_components = deepcopy(self.idx_components)
        self.input_params_dict = input_params_dict

        # spatial components
        contours = caiman_get_contours(self.cnmA[:, self.idx_components], self.dims)
        # if dfof:
        #     temporal_components = cnmC
        # else:
        #     temporal_components = cnmC[idx_components]
        self.input_params_dict = self.input_params_dict
        num_components = len(self.idx_components)

        if calc_raw_min_max:
            img = self.vi.viewer.workEnv.imgdata.seq.T

        self.ui.radioButton_curve_data.setChecked(True)

        for ix in range(num_components):
            self.vi.viewer.status_bar_label.showMessage('Please wait, adding component #: '
                                                        + str(ix) + ' / ' + str(num_components))

            curve_data = self.cnmC[self.idx_components[ix]]
            contour = contours[ix]

            if calc_raw_min_max:
                # Get a binary mask
                mask = self.cnmA[:, self.idx_components[ix]].toarray().reshape(self.dims, order='F') > 0
                # mask3d = np.array((mask,) * curve_data.shape[0])

                max_ix = curve_data.argmax()
                min_ix = curve_data.argmin()

                array_at_max = img[max_ix, :, :].copy()
                array_at_max = array_at_max[mask]

                array_at_min = img[min_ix, :, :].copy()
                array_at_min = array_at_min[mask]

                raw_min_max = self.get_raw_min_max(array_at_max=array_at_max,
                                                   array_at_min=array_at_min)

            else:
                raw_min_max = None

            roi = CNMFROI(curve_plot_item=self.get_plot_item(),
                          view_box=self.vi.viewer.getView(),
                          cnmf_idx=self.idx_components[ix],
                          curve_data=curve_data,
                          contour=contour,
                          raw_min_max=raw_min_max,
                          dfof_data=self.cnm_dfof[ix] if (self.cnm_dfof is not None) else None,
                          spike_data=self.cnmS[ix])

            self.roi_list.append(roi)

        if calc_raw_min_max:
            del img

        self.roi_list.reindex_colormap()
        self.vi.viewer.status_bar_label.showMessage('Finished adding all components!')

    def get_raw_min_max(self, array_at_max, array_at_min):
        a_size = array_at_max.size
        p5 = int(a_size * 0.05)
        p10 = p5 * 2
        p25 = p5 * 5

        out = {}

        for a, r in zip((array_at_max, array_at_min), ('raw_max', 'raw_min')):
            out[r] = {'top_5': self.get_raw_mean(a, min(5, a_size)),
                      'top_10': self.get_raw_mean(a, min(10, a_size)),
                      'top_5p': self.get_raw_mean(a, p5),
                      'top_10p': self.get_raw_mean(a, p10),
                      'top_25p': self.get_raw_mean(a, p25),
                      'full_mean': a.mean()
                      }
        return out

    def get_raw_mean(self, array, num_items):
        return np.partition(array, -num_items)[-num_items:].mean()

    def add_roi(self):
        """Not implemented, uses add_all_components to import all ROIs instead"""
        raise NotImplementedError('Not implemented for CNMFE ROIs')

    def restore_from_states(self, states: dict):
        """Restore from states, such as when these ROIs are saved with a Project Sample"""
        super(ManagerCNMFROI, self).restore_from_states(states)
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()

        if 'cnmf_data_dict' in states.keys():
            self.cnmf_data_dict = states['cnmf_data_dict']
        else:
            self.cnmf_data_dict = None

        for state in states['states']:
            roi = CNMFROI.from_state(self.get_plot_item(), self.vi.viewer.getView(), state)
            self.roi_list.append(roi)
        self.input_params_dict = states['input_params_cnmfe']
        self.cnmA = states['cnmf_output']['cnmA']
        self.cnmb = states['cnmf_output']['cnmb']
        self.cnmC = states['cnmf_output']['cnmC']
        self.cnm_f = states['cnmf_output']['cnm_f']
        self.cnmYrA = states['cnmf_output']['cnmYrA']
        self.idx_components = states['cnmf_output']['idx_components']
        self.orig_idx_components = states['cnmf_output']['orig_idx_components']
        self.roi_list.reindex_colormap()

    def get_all_states(self) -> dict:
        """Get all states so that they can be restored"""
        if not hasattr(self, 'roi_list'):
            self.create_roi_list()
        states = super(ManagerCNMFROI, self).get_all_states()

        # If the user has manually deleted some ROIs
        new_idx_components = np.array([roi.cnmf_idx for roi in self.roi_list], dtype=np.int64)

        # Store the actual cnmf attributes as well.
        input_dict = {'input_params_cnmfe': self.input_params_dict,
                      'cnmf_data_dict': self.cnmf_data_dict,
                      'cnmf_output':
                          {
                              'cnmA': self.cnmA,
                              'cnmb': self.cnmb,
                              'cnmC': self.cnmC,
                              'cnm_f': self.cnm_f,
                              'cnmYrA': self.cnmYrA,
                              'idx_components': new_idx_components,
                              'orig_idx_components': self.orig_idx_components
                          }
                      }

        states.update(input_dict)
        return states

    def update_idx_components(self, ix: int):
        """Update idx_components if the user manually delete an ROI"""
        roi = self.roi_list[self.roi_list.current_index]
        self.idx_components = np.delete(self.idx_components, np.where(self.idx_components == roi.cnmf_idx)[0])

    def set_spot_size(self, size: int):
        for roi in self.roi_list:
            roi.get_roi_graphics_object().setSize(size)
            roi.spot_size = size



import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from ...utils import get_colormap
from ..base import BasePlotWidget
from .controls import *
from ....common.qdialogs import present_exceptions
from ....common import get_window_manager
from ....common.configuration import console_history_path
from ....analysis import Transmission
from typing import *
from collections import OrderedDict
from traceback import format_exc
from ..scatter.scatter_plot import CentralWidget
from ....pyqtgraphCore import mkBrush
from ....pyqtgraphCore.console.Console import ConsoleWidget
import os


class LowDimData:
    def __init__(self):
        self.input_data: np.ndarray = np.empty(0)
        self.low_dim_data: np.ndarray = np.empty(0)

        self.labels: np.ndarray = np.empty(0, dtype='<U16')

        self.params: dict = None

        self.pca: PCA = None
        self.lda: LinearDiscriminantAnalysis = None
        
        self.lda_means: np.ndarray = np.empty(0)
        self.lda_covariance: np.ndarray = np.empty(0)
        self.lda_decision_function: np.ndarray = np.empty(0)

    def __del__(self):
        self.clear()

    def clear(self):  # in case the user does weird stuff in the console, should implement some forced garbage collection
        del self.input_data
        del self.low_dim_data

        del self.labels

        del self.params

        del self.pca
        del self.lda

        del self.lda_means
        del self.lda_covariance
        del self.lda_decision_function

        self.input_data: np.ndarray = np.empty(0)
        self.low_dim_data: np.ndarray = np.empty(0)

        self.labels: np.ndarray = np.empty(0, dtype='<U16')

        self.params: dict = None

        self.pca: PCA = None
        self.lda: LinearDiscriminantAnalysis = None

        self.lda_means: np.ndarray = np.empty(0)
        self.lda_covariance: np.ndarray = np.empty(0)
        self.lda_decision_function: np.ndarray = np.empty(0)

    def compute(
            self,
            sample_df: pd.DataFrame,
            data_column: str,
            stimulus_type: str,
            method: str,
            method_kwargs: dict = None,
            use_scaler: bool = False,
            scaler_method: str = None,

    ):
        self.params = {
            'data_column': data_column,
            'stimulus_type': stimulus_type,
            'method': method,
            'method_kwargs': method_kwargs,
            'use_scaler': use_scaler,
            'scaler_method': scaler_method,
        }

        sample_df = sample_df.reset_index(drop=True)

        # Stimulus mapping dataframe
        stim_dataframe = sample_df.iloc[0]['stim_maps'][0][0][stimulus_type]

        # create an array where each frame is labelled with the stimulus name
        self.labels = np.empty(
            shape=(sample_df[data_column].iloc[0].shape),  # shape is (n_frames,)
            dtype=f'<U{stim_dataframe.name.apply(len).max()}'  # just unicode with length of longest stimulus name str
        )

        # fill the array so each frame is labelled with the stimulus name for that frame
        for i, s in stim_dataframe.sort_values(by=['start'], ascending=True).iterrows():
            self.labels[int(s['start']):int(s['end'])] = s['name']

        self.input_data = np.vstack(sample_df[data_column].values).T

        if use_scaler:
            scaler = getattr(preprocessing, scaler_method)()
            X = scaler.fit_transform(self.input_data)
        else:
            X = self.input_data

        if method == 'PCA':
            self.pca = PCA(**method_kwargs)
            self.low_dim_data = self.pca.fit_transform(X)

        elif method == 'LDA':
            if 'store_covariance' in self.params['method_kwargs'].keys():
                store_covariance = self.params['method_kwargs'].pop('method_kwargs')

            elif 'solver' in self.params['method_kwargs'].keys():
                store_covariance = True if self.params['method_kwargs']['solver'] not in ['lsqr', 'eigen'] else False

            else:
                store_covariance=False

            lda = LinearDiscriminantAnalysis(**method_kwargs, store_covariance=store_covariance)
            self.low_dim_data = lda.fit_transform(X, self.labels)

            self.lda_means = lda.means_
            if hasattr(lda, 'covariance_'):
                self.lda_covariance = lda.covariance_
            self.lda_decision_function = lda.decision_function(X)

        return self

    def to_json_dict(self) -> dict:
        return \
            {
                'input_data': self.input_data.tolist(),
                'low_dim_data': self.low_dim_data.tolist(),
                'labels': self.labels.tolist(),
                'params': self.params,
                'lda_means': self.lda_means.tolist(),
                'lda_covariance': self.lda_covariance.tolist(),
                'lda_decision_function': self.lda_decision_function.tolist()
            }

    def from_json_dict(self, d: dict):
        attrs = list(d.keys())
        for attr in attrs:
            val = d[attr]
            if isinstance(val, list):
                val = np.array(val)

            setattr(self, attr, val)


class ControlDock(QtWidgets.QDockWidget):
    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, parent=parent)
        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.listWidgetStimulusColormap.set_cmap('tab10')

        self.ui.comboBoxMethod.addItems(['PCA', 'LDA'])

        self.ui.comboBoxPreprocessScaler.addItems(
            [
                'StandardScaler',
                'MinMaxScaler',
                'MaxAbsScaler',
                'PowerTransformer',
                'RobustScaler'
            ]
        )

        self.ui.toolBox.setCurrentIndex(0)

    @present_exceptions(
        'Cannot set params',
        'You probably have not formatted your method params correctly.\n'
        'They must be formatted as keyword arguments.'
    )
    def get_params(self) -> dict:
        _kwargs = self.ui.plainTextEditMethodParams.toPlainText()
        method_kwargs = eval(f"dict({_kwargs})")
        return \
            {
                'data_column': self.ui.comboBoxDataColumn.currentText(),
                'stimulus_type': self.ui.comboBoxStimulusMapping.currentText(),
                'method': self.ui.comboBoxMethod.currentText(),
                'method_kwargs': method_kwargs,
                'use_scaler': self.ui.checkBoxUsePreprocessScaler.isChecked(),
                'scaler_method': self.ui.comboBoxPreprocessScaler.currentText(),
            }


class NeuralDecomposePlot(QtWidgets.QMainWindow, BasePlotWidget):
    drop_opts = []

    def __init__(self):
        QtWidgets.QMainWindow.__init__(self, parent=None)
        BasePlotWidget.__init__(self)

        self.setWindowTitle('Neural Activity Decomposition')

        self.control_widget = ControlDock(self)
        self.addDockWidget(
            QtCore.Qt.LeftDockWidgetArea, self.control_widget
        )

        self.data: Dict[str, LowDimData] = None

        self.control_widget.ui.pushButtonComputeAllSamples.clicked.connect(lambda: self.compute())
        self.control_widget.ui.listWidgetComputedSamples.currentRowChanged.connect(self.update_plot)

        self.block_signals_list = [self.control_widget.ui.listWidgetComputedSamples]

        self.central_widget = CentralWidget(self)
        self.setCentralWidget(self.central_widget)

        self.plot_variant = self.central_widget.plot_variant
        self.alpha = 0.65
        self.spot_size = 10

        self.previous_point_ix: int = None
        self.previous_point_pen = None

        self.viewer_window = None

        cmd_history_file = os.path.join(console_history_path, 'neural_decompose.pik')

        ns = {'this': self}

        txt = ["Namespaces",
               "self as 'this'",
               ]

        txt = "\n".join(txt)

        self.console = ConsoleWidget(parent=self, namespace=ns, text=txt, historyFile=cmd_history_file)

        self.dockConsole = QtWidgets.QDockWidget(self)
        self.dockConsole.setWindowTitle('Console')
        self.dockConsole.setFeatures(
            QtWidgets.QDockWidget.DockWidgetFloatable | QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockConsole.setWidget(self.console)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.dockConsole)

    def set_input(self, transmission: Transmission):
        if self._transmission is not None:
            if QtWidgets.QMessageBox.warning(
                self,
                'Overwrite input?',
                'The input data to this plot node has changed.\n'
                'Do you want to load the new input data, this will clear all current data!'
            ) == QtWidgets.QMessageBox.No:
                return
            else:
                self.clear_data()

        super(NeuralDecomposePlot, self).set_input(transmission)
        self.control_widget.ui.comboBoxStimulusMapping.clear()
        self.control_widget.ui.comboBoxStimulusMapping.addItems(
            self.transmission.STIM_DEFS
        )

    def fill_control_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        self.control_widget.ui.comboBoxDataColumn.clear()
        self.control_widget.ui.comboBoxDataColumn.addItems(data_columns)

    @present_exceptions(
        'Error computing decomposition',
        'There is probably an issue with a sample(s) or your parameters'
    )
    @BasePlotWidget.signal_blocker
    def compute(self):
        if self.data is not None:
            self.clear_data()

        self.control_widget.ui.listWidgetComputedSamples.clear()

        self.data = dict()

        self.control_widget.ui.progressBar.setValue(0)

        params = self.control_widget.get_params()
        samples = self.transmission.df['SampleID'].unique()
        n_samples = samples.size

        for i, sid in enumerate(samples):
            sample_df = self.transmission.df[self.transmission.df['SampleID'] == sid]

            try:
                self.data[sid] = LowDimData().compute(sample_df, **params)
            except:
                raise ValueError(
                    f'Exception encountered for the following sample:'
                    f'\n\t{sid}\n'
                    f'{format_exc()}'
                )

            self.control_widget.ui.progressBar.setValue(int((i / n_samples) * 100))

        self.control_widget.ui.progressBar.setValue(100)

        self.control_widget.ui.listWidgetComputedSamples.addItems(
            samples.tolist()
        )

        self.control_widget.ui.toolBox.setCurrentIndex(1)

    @BasePlotWidget.signal_blocker
    def clear_data(self):
        ks = list(self.data.keys())
        for k in ks:
            print('Clearing data, please wait...')
            del self.data[k]

        self.data = None

        self.control_widget.ui.listWidgetComputedSamples.clear()

    def open_viewer(self):
        """
        Open a sample in the viewer
        """
        if self.viewer_window is not None:
            self.viewer_window.vi.viewer.sigTimeChanged.diconnect()
            self.viewer_window.close()

        w = get_window_manager().get_new_viewer_window()
        w.open_from_dataframe(proj_path=self.proj_path, row=self.row)

        self.viewer_window = w
        self.viewer_window.vi.viewer.sigTimeChanged.connect(self.highlight_point)

    def highlight_point(self, ind: tuple):
        i = ind[0]

        if self.previous_point_ix is not None:
            # self.plot_variant.plot.points()[self.previous_point_ix].resetPen()
            self.plot_variant.plot.points()[self.previous_point_ix].setPen(self.previous_point_pen)
            self.plot_variant.plot.points()[self.previous_point_ix].setSize(self.spot_size)

        self.previous_point_ix = i
        self.previous_point_pen = self.plot_variant.plot.points()[i].pen()

        # sid = self.control_widget.ui.listWidgetComputedSamples.currentItem().text()
        self.plot_variant.plot.points()[i].setPen('w')
        self.plot_variant.plot.points()[i].setSize(25)

    @present_exceptions('Plot error')
    def update_plot(self, *args, **kwargs):
        self.plot_variant.clear()

        sid = self.control_widget.ui.listWidgetComputedSamples.currentItem().text()

        low_dim_data = self.data[sid]

        xs = low_dim_data.low_dim_data[:, 0]
        ys = low_dim_data.low_dim_data[:, 1]

        colors_map = get_colormap(low_dim_data.labels,
                                  self.control_widget.ui.listWidgetStimulusColormap.get_cmap(),
                                  output='pyqt', alpha=0.7)

        colors = list(map(colors_map.get, low_dim_data.labels))
        brushes_list = list(map(mkBrush, colors))

        self.plot_variant.add_data(
            xs, ys,
            uuid_series=np.arange(0, low_dim_data.low_dim_data.shape[0]),  # use the indices to identify the points
            color=brushes_list,
            size=self.spot_size
        )

        self.plot_variant.set_legend(colors_map)

    def get_plot_opts(self, drop: bool) -> dict:
        return {
            'data': {k: v.to_json_dict() for k, v in self.data.items()},
        }

    def set_plot_opts(self, opts: dict):
        if self.data is not None:
            self.clear_data()

        self.control_widget.ui.listWidgetComputedSamples.clear()
        self.control_widget.ui.listWidgetComputedSamples.addItems(opts['data'].keys())

        ks = list(self.data.keys())

        self.data = {sid: LowDimData().from_json_dict(opts['data'][sid]) for sid in ks}

    def save_plot(self, path):
        t = self.transmission.copy()

        sid = self.transmission.df['SampleID'].unique()[0]
        params = self.data[sid].params

        self.transmission.history_trace.add_operation('all', operation='neural_decompose', parameters=params)
        super(NeuralDecomposePlot, self).save_plot(path)

        self.transmission = t



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on June 26 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007

"""

from PyQt5 import QtCore, QtGui, QtWidgets
from ....pyqtgraphCore.widgets.ColorButton import ColorButton
import pandas as pd


class Row(QtWidgets.QWidget):
    def __init__(self, pd_series=None, label=None):
        """
        :type pd_series: pd.Series
        :type label: str
        """
        QtWidgets.QWidget.__init__(self)

        self.hlayout = QtWidgets.QHBoxLayout()

        if label is not None:
            self.qlabel = QtWidgets.QLabel(self)
            self.qlabel.setText(label)
            self.hlayout.addWidget(self.qlabel)
        else:
            self.qlabel = QtWidgets.QLabel(self)

        self.name = QtWidgets.QLineEdit(self)
        self.name.setPlaceholderText('Stimulus Name')
        self.name.setMaxLength(32)
        self.hlayout.addWidget(self.name)

        self.start = QtWidgets.QLineEdit(self)
        self.start.setValidator(QtGui.QDoubleValidator())
        self.start.setPlaceholderText('Start time')
        self.hlayout.addWidget(self.start)

        self.end = QtWidgets.QLineEdit(self)
        self.end.setPlaceholderText('End time')
        self.end.setValidator(QtGui.QDoubleValidator())
        self.hlayout.addWidget(self.end)

        self.color_btn = ColorButton(self)
        self.hlayout.addWidget(self.color_btn)

        self.btn_remove = QtWidgets.QPushButton(self)
        self.btn_remove.setText('Remove stim')
        self.hlayout.addWidget(self.btn_remove)

        if pd_series is not None:
            self.set_series(pd_series)

    def set_series(self, pd_series: pd.Series):
        self.name.setText(pd_series['name'])
        self.start.setText(str(pd_series['start']))
        self.end.setText(str(pd_series['end']))
        self.color_btn.setColor(pd_series['color'])

    def get_dict(self) -> dict:
        d = {'name': self.name.text(),
             'start': float(self.start.text()),
             'end': float(self.end.text()),
             'color': self.color_btn.color(mode='byte')
             }
        return d

    def delete(self):
        self.qlabel.deleteLater()
        self.name.deleteLater()
        self.start.deleteLater()
        self.end.deleteLater()
        self.btn_remove.deleteLater()
        self.color_btn.deleteLater()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 13 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ...common.qdialogs import *
from ..core.common import ViewerUtils
from ..core import organize_metadata
from inspect import getmembers, isfunction
from .pytemplates.tiff_io_pytemplate import *
from ..core.viewer_work_environment import ViewerWorkEnv
import os
from functools import partial


class ModuleGUI(QtWidgets.QDockWidget):
    load_methods = ['asarray', 'asarray-multi', 'imread']

    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)
        QtWidgets.QDockWidget.__init__(self, parent)
        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.btnSelectTiff.clicked.connect(self.select_tiff_file)
        self.ui.btnSelectMetaFile.clicked.connect(self.select_meta_file)
        self.ui.btnLoadIntoWorkEnv.clicked.connect(self.load_tiff_file_slot)

        self.ui.radioButtonAsArray.clicked.connect(partial(self.set_load_method, 'asarray'))
        self.ui.radioButtonAsArrayMulti.clicked.connect(partial(self.set_load_method, 'asarray-multi'))
        self.ui.radioButtonImread.clicked.connect(partial(self.set_load_method, 'imread'))

        self._tiff_file_path = None
        self._meta_file_path = None
        self._load_method = None

        # all available meta data loaders
        meta_funcs = [f for f in getmembers(organize_metadata) if isfunction(f[1])]

        self.meta_loaders = dict(
            zip(
                [f[0] for f in meta_funcs],         # function name
                [f[1].__doc__ for f in meta_funcs]  # function extension from docstring
            )
        )
        self.ui.listWidget_meta_data_loader.addItems(self.meta_loaders.keys())
        self.ui.listWidget_meta_data_loader.currentItemChanged.connect(self.check_meta_path)

        self.meta_loader = None

    def file_sink(self, path):
        self.tiff_file_path = path

    @property
    def tiff_file_path(self) -> str:
        return self._tiff_file_path

    @tiff_file_path.setter
    def tiff_file_path(self, path: str):
        self._tiff_file_path = path
        self.ui.labelFileTiff.setText(self.tiff_file_path)

    @property
    def meta_file_path(self) -> str:
        return self._meta_file_path

    @meta_file_path.setter
    def meta_file_path(self, path: str):
        self._meta_file_path = path
        self.ui.labelFileMeta.setText(self.meta_file_path)

    def get_load_method(self) -> str:
        if self._load_method is None:
            raise ValueError('Load method not set.')
        else:
            return self._load_method

    def set_load_method(self, method: str):
        if method not in self.load_methods:
            raise ValueError('Must set one of the following methods: imread, asarray, asarray-multi.')
        self._load_method = method

    @use_open_file_dialog('Choose tiff file', '', ['*.tiff', '*.tif', '*.btf'])
    def select_tiff_file(self, path, *args, **kwargs):
        self.tiff_file_path = path
        self.check_meta_path()

    def check_meta_path(self) -> bool:
        """
        check if a file exists with the same name
        and the meta data extension specified by the
        selected meta data format
        """
        bn = os.path.basename(self.tiff_file_path)
        dir_path = os.path.dirname(self.tiff_file_path)
        filename = os.path.join(dir_path, os.path.splitext(bn)[0])

        if self.ui.listWidget_meta_data_loader.currentItem() is None:
            return False

        meta_loader = self.ui.listWidget_meta_data_loader.currentItem().text()

        meta_ext = self.meta_loaders[meta_loader]

        self.meta_loader = meta_loader

        meta_path = f'{filename}{meta_ext}'

        if os.path.isfile(meta_path):
            self.meta_file_path = meta_path
            self.ui.labelFileMeta.setText(meta_path)
            return True
        else:
            self.meta_file_path = ''
            return False

    @use_open_file_dialog('Choose meta data file', '')
    def select_meta_file(self, path, *args, **kwargs):
        if self.meta_loader is not None:
            ext = self.meta_loaders[self.meta_loader]
            if not path.endswith(ext):
                QtWidgets.QMessageBox.information(
                    self.parent(),
                    f'File extension mismatch',
                    f'The file extension of the chosen meta data file does not match '
                    f'the expected file extension <{ext}> for the format <{self.meta_loader}>\n\n'
                    f'You can still try to load the meta data from this file.',
                    QtWidgets.QMessageBox.Ok
                )

        self.meta_file_path = path

    @present_exceptions('Could not open file(s)', 'The following error occurred while trying to open the image file.')
    def load_tiff_file_slot(self, *args, **kwargs):
        tiff_path = self.tiff_file_path
        if not tiff_path:
            raise ValueError('Nothing to load, you must choose a file to load.')

        method = self.get_load_method()

        if self.ui.radioButton_axes_custom.isChecked():
            axes_order = self.ui.lineEdit_axes_custom.text()
        else:
            axes_order = None

        if self.ui.groupBox_meta_loader.isChecked():
            if self.meta_file_path is None:
                raise AttributeError(
                    "You have not chosen a meta data file.\n"
                    "Uncheck 'Load meta data' if you do not wish to load "
                    "meta data through this module"
                )

            if self.meta_loader is None:
                raise AttributeError(
                    "You have not selected a meta data format"
                )

            meta_path = self.meta_file_path
            meta_format = self.meta_loader
        else:
            meta_path = None
            meta_format = None

        self.load(tiff_path, method, axes_order, meta_path, meta_format=meta_format)

    def load(
            self,
            tiff_path: str,
            method: str,
            axes_order: Optional[str] = None,
            meta_path: Optional[str] = None,
            meta_format: Optional[str] = None
    ) -> ViewerWorkEnv:
        """
        Load a tiff file along with associated meta data

        :param tiff_path:       path to the tiff file
        :param meta_path:       path to the json meta data file

        :param method:          one of "asarray", "asarray-multi", or "imread"
                                    "asarray" and "asarray-multi" uses :meth:`tifffile.asarray`
                                    "asarray-multi" is for multi-page tiffs
                                    "imread" uses :meth:`tifffile.imread`

        :param axes_order:      axes order, examples: txy, xyt, tzxy, xytz etc.
        :param meta_format:     name of function from viewer.core.organize_meta that should be used
                                to organize the meta data.
        """

        self.vi.viewer.status_bar_label.showMessage('Please wait, loading tiff file, this may take a few minutes...')

        if not self.vi.discard_workEnv():
            return

        self.vi.viewer.workEnv = ViewerWorkEnv.from_tiff(
            path=tiff_path,
            method=method,
            meta_path=meta_path,
            axes_order=axes_order,
            meta_format=meta_format
        )
        self.vi.update_workEnv()
        self.vi.enable_ui(True)
        self.vi.viewer.ui.label_curr_img_seq_name.setText(os.path.basename(tiff_path))
        self.vi.viewer.status_bar_label.showMessage('File loaded into work environment!')

        return self.vi.viewer.workEnv



# from tslearn.metrics import dtw
from scipy.fftpack import rfft, irfft
import numpy as np
from multiprocessing import Pool, Manager
from PyQt5 import QtCore, QtWidgets
from functools import partial
from ...plotting.variants.timeseries import TimeseriesPlot
import traceback
from os import environ
try:
    n_processes = environ['n_processes']
except:
    n_processes = 4


def single_curve(steps: np.ndarray, interpolate: bool, q: Manager().Queue = None,
                 raw_curve: np.ndarray = None, rf_curve: np.ndarray = None) -> list:
    """Compute for a single curve and return a 1D array of distances where each element is an incremental step.
    Must provide either raw curve or rfft of curve"""
    if (raw_curve is None) and (rf_curve is None):
        raise ValueError('Must provide either raw curve or rfft of curve')

    if rf_curve is None:
        rf = rfft(raw_curve)
    else:
        rf = rf_curve

    if not interpolate:
        rval = _single_curve(raw_curve, rf, steps)

    else:
        rval = _single_curve_interp(raw_curve, rf, steps)

    if q is not None:
        q.put(1)

    return rval


def _single_curve(raw_curve, rf, steps) -> list:
    dtw_dists = []
    for step in steps:
        irf = irfft(rf[:step])
        dist = dtw(raw_curve, irf)
        dtw_dists.append(dist)

    return dtw_dists


def _single_curve_interp(raw_curve, rf, steps) -> list:
    dtw_dists = []
    x = np.arange(0, len(raw_curve))
    for step in steps:
        irf = irfft(rf[:step])
        xp = np.linspace(0, raw_curve.size, step)
        interp = np.interp(x, xp, fp=irf)

        dist = dtw(raw_curve, interp)
        dtw_dists.append(dist)

    return dtw_dists


def curves_2d(raw_curves: np.ndarray, start: int, step: int, interpolate: bool) -> np.ndarray:
    """Compute for a 2D array of curves
    :param raw_curves: (n_curves, y_values of curve)"""

    p = Pool(n_processes)
    fft_curves = rfft(raw_curves)

    dists = p.starmap(partial(single_curve, start, step, interpolate), zip(raw_curves, fft_curves))


class _Signals(QtCore.QObject):
    update_progress_bar = QtCore.pyqtSignal()
    result = QtCore.pyqtSignal(np.ndarray)
    finished = QtCore.pyqtSignal()
    error = QtCore.pyqtSignal(str)


class Widget(QtWidgets.QWidget):
    sig_set_progress_bar = QtCore.pyqtSignal(float)

    def __init__(self):
        QtWidgets.QWidget.__init__(self)

        layout = QtWidgets.QVBoxLayout()

        abort_button = QtWidgets.QPushButton()
        abort_button.setText('Abort')
        abort_button.clicked.connect(self.abort_requested)
        layout.addWidget(abort_button)

        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setValue(0)
        layout.addWidget(self.progress_bar)

        self.label = QtWidgets.QLabel()
        self.label.setText('')
        layout.addWidget(self.label)

        self.plot = TimeseriesPlot()
        self.plot.hide()
        layout.addWidget(self.plot)

        self.setLayout(layout)

    def abort_requested(self):
        if QtWidgets.QMessageBox.question(self, 'Abort processes?',
                                         'Are you sure you want to abort?') == QtWidgets.QMessageBox.No:
            return
        self.runner.p.terminate()
        self.thread_pool.cancel(self.runner)

    @QtCore.pyqtSlot()
    def update_progress_bar(self):
        self.num_processed += 1
        v = int(self.num_processed / self.num_curves)
        print(v)
        self.progress_bar.setValue(v * 100)

    def start(self, start: int, step: int, interpolate: bool,
              raw_curves: np.ndarray):

        self.num_curves = raw_curves.shape[0]
        self.num_processed = 0

        self.runner = _Runner(start, step, interpolate, raw_curves)

        self.runner.signals.update_progress_bar.connect(self.update_progress_bar)
        self.runner.signals.result.connect(self.set_plot)
        self.runner.signals.finished.connect(lambda: self.label.setText('Done!'))
        self.runner.signals.finished.connect(lambda: self.progress_bar.setValue(100))
        self.runner.signals.error.connect(lambda e: QtWidgets.QMessageBox.warning(None, 'Error', e))

        self.thread_pool = QtCore.QThreadPool()
        self.thread_pool.setMaxThreadCount(1)
        self.thread_pool.start(self.runner)

    # @QtCore.pyqtSlot(np.ndarray)
    def set_plot(self, result: np.ndarray):
        self.plot.clear()
        self.plot.set(result)
        self.plot.show()


class _Runner(QtCore.QRunnable):
    def __init__(self, start: int, step: int, interpolate: bool,
              raw_curves: np.ndarray):
        super(_Runner, self).__init__()

        assert isinstance(raw_curves, np.ndarray)

        self.signals = _Signals()

        self.start_param = start
        self.step_param = step
        self.interpolate_param = interpolate
        self.raw_curves = raw_curves

    def run(self):
        try:
            self.p = Pool(n_processes)
            rf_curve = rfft(self.raw_curves)
            q = Manager().Queue()

            steps = np.arange(self.start_param, len(self.raw_curves[0, :]), self.step_param)

            self.dists = self.p.starmap(partial(single_curve, steps, self.interpolate_param, q), zip(self.raw_curves, rf_curve))

            num_finished = 0
            num_curves = self.raw_curves.shape[0]

            # while num_finished < num_curves:
            #     sleep(1)
            #     if q.qsize() > 0:
            #         q.get()
            #         self.signals.update_progress_bar.emit()

            dists_array = np.array(self.dists)

        except:
            self.signals.error.emit(traceback.format_exc())
        else:
            self.signals.result.emit(dists_array)
        finally:
            self.p.terminate()
            self.signals.finished.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on April 23 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..core.common import ViewerUtils
from .pytemplates.cnmf_pytemplate import *
from ...common import get_window_manager
from ...common.qdialogs import *
from ...common.utils import HdfTools
from uuid import UUID
from shutil import copy
import os
from copy import deepcopy
import logging


logger = logging.getLogger()


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)

        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.btnAddToBatchCNMF.clicked.connect(lambda: self.add_to_batch())

    @present_exceptions()
    def get_params(self, *args, group_params: bool = False) -> dict:
        """
        Get a dict of the set parameters.
        If the work environment was loaded from a motion correction batch item it put the bord_px in the dict.
        Doesn't use any arguments

        :return: parameters dict
        :rtype: dict
        """
        if self.vi.viewer.workEnv.imgdata.meta['fps'] == 0:
            raise KeyError('No framerate for current image sequence!',
                           'You must set a framerate for the current image sequence. '
                           'You can set it manually in the console like this:\nget_meta()["fps"] = <framerate>')

        # Get width of border that is NaN, usually happens due ot motion correction
        history_trace = self.vi.viewer.workEnv.history_trace
        try:
            bord_px = next(d for ix, d in enumerate(history_trace) if 'caiman_motion_correction' in d)['caiman_motion_correction']['bord_px']
        except StopIteration:
            bord_px = 0

        rf = self.ui.spinBoxRf.value()

        # CNMF kwargs
        cnmf_kwargs = \
            {
                'p': self.ui.spinBoxP.value(),
                'nb': self.ui.spinBoxnb.value(),
                'merge_thresh': self.ui.doubleSpinBoxMergeThresh.value(),
                'rf': rf if not self.ui.checkBoxRfNone.isChecked() else None,
                'stride': self.ui.spinBoxStrideCNMF.value(),
                'K': self.ui.spinBoxK.value(),
                'gSig': [
                            self.ui.spinBox_gSig_x.value(),
                            self.ui.spinBox_gSig_y.value()
                        ],
                'ssub': self.ui.spinBox_ssub.value(),
                'tsub': self.ui.spinBox_tsub.value(),
                'method_init': self.ui.comboBox_method_init.currentText(),
                'border_pix': bord_px,
                'fr': self.vi.viewer.workEnv.imgdata.meta['fps']
            }

        # Any additional cnmf kwargs set in the text entry
        if self.ui.groupBox_cnmf_kwargs.isChecked():
            try:
                _kwargs = self.ui.plainTextEdit_cnmf_kwargs.toPlainText()
                cnmf_kwargs_add = eval(f"dict({_kwargs})")
                cnmf_kwargs.update(cnmf_kwargs_add)
            except:
                raise ValueError("CNMF kwargs not formatted properly.")

        # Component evaluation kwargs
        eval_kwargs = \
            {
                'min_SNR': self.ui.doubleSpinBoxMinSNR.value(),
                'rval_thr': self.ui.doubleSpinBoxRvalThr.value(),
                'use_cnn': self.ui.checkBoxUseCNN.isChecked(),
                'min_cnn_thr': self.ui.doubleSpinBoxCNNThr.value(),
                'cnn_lowest': self.ui.doubleSpinBox_cnn_lowest.value(),
                'decay_time': self.ui.spinBoxDecayTime.value(),
                'fr': self.vi.viewer.workEnv.imgdata.meta['fps']
            }

        # Any additional eval kwargs set in the text entry
        if self.ui.groupBox_eval_kwargs.isChecked():
            try:
                _kwargs = self.ui.plainTextEdit_eval_kwargs.toPlainText()
                eval_kwargs_add = eval(f"dict{_kwargs}")
                eval_kwargs.update(eval_kwargs_add)
            except:
                raise ValueError("Evaluation kwargs not formatted properly.")

        if self.vi.viewer.workEnv.imgdata.ndim == 4:
            is_3d = True
        else:
            is_3d = False

        # Make the output dict
        d = \
            {
                'item_name': self.ui.lineEdName.text(),
                'refit': self.ui.checkBoxRefit.isChecked(),
                'border_pix': bord_px,
                'is_3d': is_3d,
                'keep_memmap': self.ui.checkBoxKeepMemmap.isChecked()
            }

        # Group the kwargs of the two parts seperately
        if group_params:
            d.update(
                {
                    'cnmf_kwargs': cnmf_kwargs,
                    'eval_kwargs': eval_kwargs
                }
            )

        # or not
        else:
            d.update(
                {
                    **cnmf_kwargs,
                    **eval_kwargs
                }
            )

        return d

    @present_exceptions()
    def add_to_batch(self, params: dict = None) -> UUID:
        """
        Add a CNMF batch item with the currently set parameters and the current work environment.
        """
        if self.vi.viewer.workEnv.isEmpty:
            QtWidgets.QMessageBox.warning(self, 'Empty work environment', 'The work environment is empty, '
                                                                          'nothing to add to batch')
            return

        input_workEnv = self.vi.viewer.workEnv

        if params is None:
            d = self.get_params(group_params=True)
        else:
            # Check that user passed dict is formatted correctly
            required_keys = ['cnmf_kwargs', 'eval_kwargs', 'item_name', 'refit', 'border_pix']
            if any([k not in params.keys() for k in required_keys]):
                raise ValueError(f'Must pass a params dict with the following keys:\n'
                                 f'{required_keys}\n'
                                 f'Please see the docs for more information.')
            d = deepcopy(params)

        name = d['item_name']
        self.vi.viewer.status_bar_label.showMessage('Please wait, adding CNMF: ' + name + ' to batch...')

        if self.ui.groupBox_seed_components.isChecked():
            seed_path = self.ui.lineEdit_seed_components_path.text()
            if not os.path.isfile(seed_path):
                raise FileNotFoundError(
                    "Seed file does not exist, check the path"
                )

            try:
                seed_params = HdfTools.load_dict(seed_path, 'data/segment_params')
            except:
                seed_params = 'unknown'

            d['use_seeds'] = True
            d['seed_params'] = seed_params
        else:
            d['use_seeds'] = False

        batch_manager = get_window_manager().get_batch_manager()
        u = batch_manager.add_item(
            module='CNMF',
            name=name,
            input_workEnv=input_workEnv,
            input_params=d,
            info=d
        )

        if u is None:
            self.vi.viewer.status_bar_label.clearMessage()
            return

        if d['use_seeds']:
            logger.info("Copying component seed file")
            copy(seed_path, os.path.join(batch_manager.batch_path, f'{u}.ain'))

        self.vi.viewer.status_bar_label.showMessage('Done adding CNMF: ' + name + ' to batch!')
        self.ui.lineEdName.clear()

        return u



import os
import glob
import importlib.util

dirname = os.path.dirname(__file__)
fname = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [os.path.basename(f)[:-3] for f in fname if os.path.isfile(f) and not f.endswith('__init__.py')]

for m in __all__:
    spec = importlib.util.spec_from_file_location(m, )




from ...common.configuration import HAS_TSLEARN

from .datapoint_tracer import DatapointTracerWidget

from .heatmap import HeatmapWidget, HeatmapSplitterWidget, HeatmapTracerWidget
from .cross_correlation import CrossCorrelationWidget
from .scatter import ScatterPlotWidget

from .beeswarms import BeeswarmPlotWindow
from .curve_plot_window import CurvePlotWindow

from .proportions import ProportionsWidget
from .space_map import SpaceMapWidget

from .stimulus_tuning import TuningCurvesWidget
from .neural_decompose import NeuralDecomposePlot

__all__ = ['DatapointTracerWidget', 'HeatmapWidget', 'HeatmapSplitterWidget',
           'HeatmapTracerWidget', 'CrossCorrelationWidget', 'ScatterPlotWidget', 'BeeswarmPlotWindow',
           'CurvePlotWindow', 'ProportionsWidget', 'SpaceMapWidget', 'TuningCurvesWidget', 'NeuralDecomposePlot']

if HAS_TSLEARN:
    from .kshape import KShapeWidget
    __all__ += ['KShapeWidget']



from sklearn.metrics import pairwise_distances
from typing import Tuple, Optional, Union
import numpy as np
from ..common import get_sys_config


def get_centerlike(cluster_members: np.ndarray, metric: Optional[Union[str, callable]] = None,
                   dist_matrix: Optional[np.ndarray] = None) -> Tuple[np.ndarray, int]:
    """
    Finds the 1D time-series within a cluster that is the most centerlike

    :param cluster_members: 2D numpy array in the form [n_samples, 1D time_series]
    :param metric:          Metric to use for pairwise distance calculation, simply passed to sklearn.metrics.pairwise_distances
    :param dist_matrix:     Distance matrix of the cluster members

    :return:                The cluster member which is most centerlike, and its index in the cluster_members array
    """
    if dist_matrix is None:
        dist_matrix = pairwise_distances(cluster_members, metric=metric,
                                         n_jobs=get_sys_config()['_MESMERIZE_N_THREADS'])

    c_ix = np.argmin(np.sum(dist_matrix, axis=0))
    c = cluster_members[c_ix, :]
    return c, c_ix


def get_cluster_radius(cluster_members: np.ndarray, metric: Optional[Union[str, callable]] = None,
                       dist_matrix: Optional[np.ndarray] = None, centerlike_index: Optional[int] = None) -> float:
    """
    Returns the cluster radius according to chosen distance metric

    :param cluster_members: 2D numpy array in the form [n_samples, 1D time_series]
    :param metric:          Metric to use for pairwise distance calculation, simply passed to sklearn.metrics.pairwise_distances
    :param dist_matrix:     Distance matrix of the cluster members
    :param centerlike_index: Index of the centerlike cluster member within the cluster_members array

    :return:                 The cluster radius, average between the most centerlike member and all other members
    """
    if dist_matrix is None:
        dist_matrix = pairwise_distances(cluster_members, metric=metric,
                                         n_jobs=get_sys_config()['_MESMERIZE_N_THREADS'])

    if centerlike_index is None:
        c, c_ix = get_centerlike(cluster_members, dist_matrix)
    else:
        c_ix = centerlike_index

    ra = np.sum(dist_matrix[c_ix, :]) / (cluster_members.shape[0] - 1)
    return ra


def davies_bouldin_score(
        data: np.ndarray, cluster_labels: np.ndarray, metric: Union[str, callable]
    ) -> Tuple[float, np.ndarray]:
    """
    Adopted from sklearn.metrics.davies_bouldin_score to use any distance metric

    :param data:            Data that was used for clustering, [n_samples, 1D time_series]
    :param metric:          Metric to use for pairwise distance calculation, simply passed to sklearn.metrics.pairwise_distances
    :param cluster_labels:  Cluster labels

    :return:                Davies Bouldin Score using EMD
    """
    radii = np.zeros(np.unique(cluster_labels).size)
    centroids = np.zeros((np.unique(cluster_labels).size, data.shape[1]))

    for i, c in enumerate(np.unique(cluster_labels)):
        members = np.take(data, np.where(cluster_labels == c)[0], axis=0)
        dist_m = pairwise_distances(members, metric=metric, n_jobs=get_sys_config()['_MESMERIZE_N_THREADS'])

        centroids[i, :], c_ix = get_centerlike(members, dist_matrix=dist_m)
        radii[i] = get_cluster_radius(members, dist_matrix=dist_m, centerlike_index=c_ix)

    centroid_distances = pairwise_distances(centroids, metric=metric, n_jobs=get_sys_config()['_MESMERIZE_N_THREADS'])

    if np.allclose(radii, 0) or np.allclose(centroid_distances, 0):
        return 0.0, None

    centroid_distances[centroid_distances == 0] = np.inf
    combined_intra_dists = radii[:, None] + radii
    combined_intra_dists = np.nan_to_num(combined_intra_dists, nan=0.0)  # for cases where a cluster has just 1 member

    scores = np.max(combined_intra_dists / centroid_distances, axis=1)

    return (np.nanmean(scores), scores)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtWidgets
from ....analysis.math.cross_correlation import compute_cc_data, CC_Data
from .control_widget_pytemplate import Ui_CrossCorrelationControls
from .. import HeatmapSplitterWidget
from ...variants import TimeseriesPlot
from ....analysis import Transmission, get_sampling_rate, get_array_size, organize_dataframe_columns
import numpy as np
from ....pyqtgraphCore import PlotDataItem, mkPen
from ..datapoint_tracer import DatapointTracerWidget, CNMFROI, ManualROI, mkColor
import pandas as pd
from ....common.qdialogs import *
from ....common.utils import HdfTools
from .stims import get_binary_stims


class ControlWidget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        self.ui = Ui_CrossCorrelationControls()
        self.ui.setupUi(self)


class CrossCorrelationWidget(HeatmapSplitterWidget):
    def __init__(self):
        HeatmapSplitterWidget.__init__(self, highlight_mode='item')
        self.control_widget = ControlWidget()
        # self.cross_corr_layout = QtWidgets.QVBoxLayout()
        self.add_to_splitter(self.control_widget)

        self.cross_corr_plot = TimeseriesPlot()
        self.control_widget.ui.verticalLayoutLeftMain.addWidget(self.cross_corr_plot)

        self.current_sample_id = None
        self.control_widget.ui.listWidgetSampleID.currentItemChanged.connect(self.set_current_sample)

        # self.main_dataframe = None
        self.transmission = None
        self.sample_list = None
        self.cc_data = None

        self.plot_variant.sig_selection_changed.connect(self.set_lineplots)

        # self.plot_widget.ax_ylabel_bar.set_axis_on()
        # self.plot_widget.ax_heatmap.get_yaxis().set_visible(True)

        self.curve_plot_1 = PlotDataItem()
        self.curve_plot_2 = PlotDataItem()

        self.control_widget.ui.graphicsViewCurve1.addItem(self.curve_plot_1)
        self.control_widget.ui.graphicsViewCurve2.addItem(self.curve_plot_2)

        self.control_widget.ui.pushButtonComputeAllData.clicked.connect(self.compute_dataframe)
        self.control_widget.ui.pushButtonExportAllData.clicked.connect(self.export_data)

        self.control_widget.ui.radioButtonMaxima.clicked.connect(self.set_heatmap)
        self.control_widget.ui.radioButtonLag.clicked.connect(self.set_heatmap)
        self.control_widget.ui.doubleSpinBoxMaximaThreshold.valueChanged.connect(self.set_heatmap)
        self.control_widget.ui.doubleSpinBoxLagThreshold.valueChanged.connect(self.set_heatmap)
        self.datapoint_tracer = DatapointTracerWidget()

        self.stimulus_labels: List[str] = []

        self.roi_2 = None

    def set_heatmap(self):
        if self.cc_data is None:
            return
        
        sample_id = self.current_sample_id

        abs_val = self.control_widget.ui.checkBoxAbsoluteValue.isChecked()

        lt = self.control_widget.ui.doubleSpinBoxLagThreshold.value()
        mt = self.control_widget.ui.doubleSpinBoxMaximaThreshold.value()

        if self.control_widget.ui.radioButtonLag.isChecked():
            opt = 'lag'
            cmap = 'brg'

        elif self.control_widget.ui.radioButtonMaxima.isChecked():
            opt = 'maxima'
            cmap = 'jet'
        else:
            raise ValueError

        plot_data = self.cc_data[sample_id].get_threshold_matrix(matrix_type=opt, lag_thr=lt, max_thr=mt, lag_thr_abs=abs_val)

        sub_df = self.transmission.df[self.transmission.df['SampleID'] == sample_id]

        labels_col = self.control_widget.ui.comboBoxLabelsColumn.currentText()
        ylabels = sub_df[labels_col].tolist() + self.stimulus_labels

        # Since it's just same curve vs. same curves in the cc_matrix
        xlabels = ylabels
                                                                                     #  lines make large matrices hard to see
        self.plot_variant.set(plot_data, cmap=cmap, ylabels=ylabels, xlabels=xlabels)#, linewidths=.5, linecolor='k')
        # self.plot_widget.ax_ylabel_bar.set_axis_off()

        self.plot_variant.plot.ax_heatmap.set_xlabel('Curve 1, magenta')
        self.plot_variant.plot.ax_heatmap.set_ylabel('Curve 2, cyan')

        self.plot_variant.draw()

    def set_lineplots(self, indices):
        if self.plot_variant.selector.multi_select_mode:
            nccs = []
            for ix in self.plot_variant.selector.multi_selection_list:
                # x = self.data[ix[0]]
                # y = self.data[ix[1]]
                i = ix[0]
                j = ix[1]
                nccs.append(self.cc_data[self.current_sample_id].ccs[i, j, :])

            if len(nccs) < 1:
                return

            a = np.vstack(nccs)

            self.curve_plot_1.clear()
            self.curve_plot_2.clear()

            xticks = self._get_xticks_linspace(nccs[0])

            self.cross_corr_plot.set(a, xticks=xticks)
        else:
            i = indices[0]
            j = indices[1]
            x = self.cc_data[self.current_sample_id].input_data[i]
            y = self.cc_data[self.current_sample_id].input_data[j]
            # x = self.curve_data[i]
            # y = self.curve_data[j]

            self.curve_plot_1.clear()
            self.curve_plot_2.clear()

            sub_df = self.transmission.df[self.transmission.df.SampleID == self.current_sample_id].reset_index(drop=True)

            if i < sub_df.uuid_curve.size:
                ux = sub_df.uuid_curve.iloc[i]
            else:
                ux = False

            if j < sub_df.uuid_curve.size:
                uy = sub_df.uuid_curve.iloc[j]
            else:
                uy = False

            if ux:
                r = sub_df[sub_df.uuid_curve == ux]
                db_id = r._BLOCK_.item()

                self.datapoint_tracer.set_widget(
                    datapoint_uuid=ux, data_column_curve=self.data_column,
                    row=r, proj_path=self.transmission.get_proj_path(),
                    history_trace=self.transmission.history_trace.get_data_block_history(db_id),
                    roi_color='m'
                )
            self.datapoint_tracer.ui.graphicsViewPlot.clear()

            if uy:
                r2 = sub_df[sub_df.uuid_curve == uy]
                self.add_second_roi_to_datapoint_tracer(r2)

            if (ux or uy):
                self.datapoint_tracer.show()

            self.control_widget.ui.lineEditCurve1UUID.setText(ux if ux else '')
            self.control_widget.ui.lineEditCurve2UUID.setText(uy if uy else '')

            self.curve_plot_1.setData(x=np.linspace(0, (len(x) / self.sampling_rate), len(x)), y=x, pen=mkPen(color='m', width=2))
            self.curve_plot_2.setData(x=np.linspace(0, (len(y) / self.sampling_rate), len(y)), y=y, pen=mkPen(color='c', width=2))

            ncc = self.cc_data[self.current_sample_id].ccs[i, j, :]

            xticks = self._get_xticks_linspace(ncc)

            self.cross_corr_plot.set_single_line(x=xticks, y=ncc)
        self.cross_corr_plot.ax.set_xlabel("lag (seconds)")

    # TODO: Just simplify the datapoint_tracer to take in multiple ROIs
    def add_second_roi_to_datapoint_tracer(self, r: pd.Series):
        if isinstance(self.roi_2, (CNMFROI, ManualROI)):
            self.roi_2.remove_from_viewer()

        roi_state = r['ROI_State'].item()
        if isinstance(roi_state, pd.Series):
            roi_state = roi_state.item()

        if roi_state['roi_type'] == 'CNMFROI':
            self.roi_2 = CNMFROI.from_state(curve_plot_item=None, view_box=self.datapoint_tracer.view, state=roi_state)
            self.roi_2.get_roi_graphics_object().setBrush(mkColor('c'))

        elif roi_state['roi_type'] == 'ManualROI':
            self.roi_2 = ManualROI.from_state(curve_plot_item=None, view_box=self.datapoint_tracer.view, state=roi_state)

        if self.roi_2 is not None:
            self.roi_2.get_roi_graphics_object().setPen(mkColor('c'))
            self.roi_2.add_to_viewer()

    def _get_xticks_linspace(self, ncc) -> np.ndarray:
            m = ncc.size
            stop = ((m / 2) / self.sampling_rate)
            start = -stop
            return np.linspace(start, stop, m)

    def set_input(self, transmission: Transmission = None):
        self.transmission = transmission
        self.transmission.df.reset_index(drop=True, inplace=True)

        cols = self.transmission.df.columns
        dcols, ccols, ucols = organize_dataframe_columns(cols)

        self.control_widget.ui.comboBoxDataColumn.clear()
        self.control_widget.ui.comboBoxDataColumn.addItems(dcols)

        self.control_widget.ui.comboBoxLabelsColumn.clear()
        self.control_widget.ui.comboBoxLabelsColumn.addItems(ccols)

        self.control_widget.ui.comboBoxStimulusType.clear()
        self.control_widget.ui.comboBoxStimulusType.addItems(
            self.transmission.STIM_DEFS
        )

        self.reset_sample_id_list_widget()
        self.control_widget.ui.listWidgetSampleID.setCurrentRow(0)

    def reset_sample_id_list_widget(self):
        self.sample_list = self.transmission.df.SampleID.unique().tolist()
        self.control_widget.ui.listWidgetSampleID.clear()
        self.control_widget.ui.listWidgetSampleID.addItems(self.sample_list)

    def set_current_sample(self):
        if self.cc_data is None:
            return
        ix = self.control_widget.ui.listWidgetSampleID.currentRow()
        self.current_sample_id = self.sample_list[ix]
        self.curve_data = np.vstack(self.transmission.df[self.transmission.df.SampleID == self.current_sample_id][self.data_column].values)
        self.set_heatmap()

    def compute_dataframe(self):
        self.data_column = self.control_widget.ui.comboBoxDataColumn.currentText()
        labels_col = self.control_widget.ui.comboBoxLabelsColumn.currentText()

        self.cc_data = dict.fromkeys(self.sample_list)

        for sample_id in self.sample_list:
            sub_df = self.transmission.df[self.transmission.df.SampleID == sample_id]

            if self.transmission.STIM_DEFS:
                stim_type = self.control_widget.ui.comboBoxStimulusType.currentText()

                stim_df = sub_df.iloc[0].stim_maps[0][0][stim_type]
                index_size = sub_df[self.data_column].values[0].size

                binary_stims_array, stim_names = get_binary_stims(
                    stim_df=stim_df,
                    index_size=index_size,
                    start_offset=0,
                    end_offset=0
                )

            data = np.vstack(sub_df[self.data_column].values)

            # append the stimulus array too
            if self.transmission.STIM_DEFS:
                data = np.vstack([data, binary_stims_array])

            r = get_sampling_rate(self.transmission)
            self.sampling_rate = r

            self.cc_data[sample_id] = compute_cc_data(data)
            self.cc_data[sample_id].lag_matrix = np.true_divide(self.cc_data[sample_id].lag_matrix, r)
            self.cc_data[sample_id].curve_uuids = np.array(list(map(str, sub_df['uuid_curve'].values))) # convert all UUIDs to str representation

            labels = sub_df[labels_col].values.astype(np.unicode)

            if self.transmission.STIM_DEFS:
                self.cc_data[sample_id].labels = np.concatenate([labels, stim_names]).astype(np.unicode)
                self.stimulus_labels = stim_names
            else:
                self.cc_data[sample_id].labels = labels
                self.stimulus_labels = []

        self.set_current_sample()

    @use_save_file_dialog('Save file as', None, '.hdf5')
    @present_exceptions()
    def export_data(self, path, *args, **kwargs):
        HdfTools.save_dict(self.cc_data, path, group='cross_corr_data')



# -*- coding: utf-8 -*-
# from . import functions
from .common import *
from ....analysis.data_types import *
from ....analysis.stimulus_extraction import StimulusExtraction
from ....plotting import TuningCurvesWidget


class ExtractStim(CtrlNode):
    """Extract portions of curves according to stimulus maps"""
    nodeName = 'ExtractStim'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Stim_Type', 'combo', {}),
                  ('start_offset', 'intSpin', {'min': -999999, 'max': 999999, 'value': 0, 'step': 1}),
                  ('end_offset', 'intSpin', {'min': -999999, 'max': 999999, 'value': 0, 'step': 1}),
                  ('zero_pos', 'combo', {'values': ['start_offset', 'stim_end', 'stim_center']}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()
        self.ctrls['Stim_Type'].setItems(self.t.STIM_DEFS)

        if self.ctrls['Apply'].isChecked() is False:
            return

        self.t = transmission.copy()

        data_column = self.ctrls['data_column'].currentText()

        stim_def = self.ctrls['Stim_Type'].currentText()
        start_offset = self.ctrls['start_offset'].value()
        end_offset = self.ctrls['end_offset'].value()
        zero_pos = self.ctrls['zero_pos'].currentText()

        self.stim_extractor = StimulusExtraction(self.t, data_column, stim_def, start_offset, end_offset, zero_pos)

        self.t = self.stim_extractor.extract()

        params = {'stim_type': stim_def,
                  'start_offset': start_offset,
                  'end_offset': end_offset,
                  'zero_pos': zero_pos
                  }

        self.t.history_trace.add_operation('all', operation='extract_stim', parameters=params)
        self.t.last_output = '_ST_CURVE'

        return self.t


class StimTuning(CtrlNode):
    """Get stimulus tuning curves"""
    nodeName = "StimTuning"
    uiTemplate = [('ShowGUI', 'button', {'text': 'Show GUI'})]

    def __init__(self, name):
        CtrlNode.__init__(self, name, terminals={'In': {'io': 'in', 'multi': True}, 'Out': {'io': 'out'}})
        self.widget = TuningCurvesWidget()
        self.widget.sig_output_changed.connect(self._send_output)
        self.ctrls['ShowGUI'].clicked.connect(self.widget.show)
        self.output_transmission = None

    def _send_output(self, t: Transmission):
        self.output_transmission = t
        self.changed()

    def process(self, output_transmission=False, **kwargs):
        if self.output_transmission:
            out = self.output_transmission
        else:
            out = self.processData(**kwargs)

        return {'Out': out}

    def processData(self, **kwargs):
        self.transmissions = kwargs['In']
        self.transmissions_list = merge_transmissions(self.transmissions)
        self.t = Transmission.merge(self.transmissions_list)
        self.widget.set_input(self.t)
        return None


class StaticDFoFo(CtrlNode):
    """Perform (F - Fo / Fo) without a rolling window to find Fo"""
    nodeName = 'StaticDFoFo'
    uiTemplate = [('data_column', 'combo', {}),
                  ('Apply', 'check', {'checked': False, 'applyBox': True})
                  ]

    def processData(self, transmission: Transmission):
        self.t = transmission
        self.set_data_column_combo_box()

        if not self.ctrls['Apply'].isChecked():
            return

        self.t = transmission.copy()

        data_column = self.ctrls['data_column'].currentText()
        output_column = '_STATIC_DF_O_F'
        params = {'data_column': data_column}

        self.t.df[output_column] = self.t.df[data_column].apply(lambda a: self._static_dfof(a))

        self.t.history_trace.add_operation(data_block_id='all', operation='static_df_o_f', parameters=params)
        self.t.last_output = output_column

        return self.t

    def _static_dfof(self, F: np.ndarray) -> np.ndarray:
        Fo = np.min(F)
        d = ((F - Fo) / Fo)
        return d


class ManualDFoF(CtrlNode):
    """Set Fo for dF/Fo using a particular time period. Useful for looking at stimulus responses"""
    nodeName = 'ManualDFoF'
    uiTemplate = [('Apply', 'check', {'checked': True, 'applyBox': True}),
                  ('data_column', 'combo', {}),
                  ('OpenGUI', 'button', {'text': 'OpenGUI'})]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './control_widget.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_KShapeControl(object):
    def setupUi(self, KShapeControl):
        KShapeControl.setObjectName("KShapeControl")
        KShapeControl.resize(524, 935)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBoxKShapeParams = QtWidgets.QGroupBox(self.dockWidgetContents)
        self.groupBoxKShapeParams.setObjectName("groupBoxKShapeParams")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.groupBoxKShapeParams)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.pushButtonReconnectFlowchartInput = QtWidgets.QPushButton(self.groupBoxKShapeParams)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(False)
        font.setWeight(50)
        self.pushButtonReconnectFlowchartInput.setFont(font)
        self.pushButtonReconnectFlowchartInput.setObjectName("pushButtonReconnectFlowchartInput")
        self.verticalLayout_7.addWidget(self.pushButtonReconnectFlowchartInput)
        self.labelConnectedToFlowchart = QtWidgets.QLabel(self.groupBoxKShapeParams)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.labelConnectedToFlowchart.setFont(font)
        self.labelConnectedToFlowchart.setAlignment(QtCore.Qt.AlignCenter)
        self.labelConnectedToFlowchart.setObjectName("labelConnectedToFlowchart")
        self.verticalLayout_7.addWidget(self.labelConnectedToFlowchart)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        self.label_8 = QtWidgets.QLabel(self.groupBoxKShapeParams)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_12.addWidget(self.label_8)
        self.comboBoxDataColumn = QtWidgets.QComboBox(self.groupBoxKShapeParams)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxDataColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxDataColumn.setSizePolicy(sizePolicy)
        self.comboBoxDataColumn.setObjectName("comboBoxDataColumn")
        self.horizontalLayout_12.addWidget(self.comboBoxDataColumn)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_12.addItem(spacerItem)
        self.verticalLayout_6.addLayout(self.horizontalLayout_12)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.groupBoxKShapeParams)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.spinBoxMaxIter = QtWidgets.QSpinBox(self.groupBoxKShapeParams)
        self.spinBoxMaxIter.setMaximum(9999)
        self.spinBoxMaxIter.setSingleStep(50)
        self.spinBoxMaxIter.setProperty("value", 300)
        self.spinBoxMaxIter.setObjectName("spinBoxMaxIter")
        self.horizontalLayout_2.addWidget(self.spinBoxMaxIter)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_6.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QtWidgets.QLabel(self.groupBoxKShapeParams)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_3.addWidget(self.label_3)
        self.spinBoxTol = QtWidgets.QSpinBox(self.groupBoxKShapeParams)
        self.spinBoxTol.setMinimum(-20)
        self.spinBoxTol.setMaximum(10)
        self.spinBoxTol.setSingleStep(1)
        self.spinBoxTol.setProperty("value", -6)
        self.spinBoxTol.setObjectName("spinBoxTol")
        self.horizontalLayout_3.addWidget(self.spinBoxTol)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_4 = QtWidgets.QLabel(self.groupBoxKShapeParams)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_4.addWidget(self.label_4)
        self.spinBoxN_init = QtWidgets.QSpinBox(self.groupBoxKShapeParams)
        self.spinBoxN_init.setMinimum(1)
        self.spinBoxN_init.setObjectName("spinBoxN_init")
        self.horizontalLayout_4.addWidget(self.spinBoxN_init)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_4)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.verticalLayout_6.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_15 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_15.setObjectName("horizontalLayout_15")
        self.label_11 = QtWidgets.QLabel(self.groupBoxKShapeParams)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_15.addWidget(self.label_11)
        self.spinBoxTrainSubsetPercentage = QtWidgets.QSpinBox(self.groupBoxKShapeParams)
        self.spinBoxTrainSubsetPercentage.setMinimum(30)
        self.spinBoxTrainSubsetPercentage.setMaximum(100)
        self.spinBoxTrainSubsetPercentage.setSingleStep(10)
        self.spinBoxTrainSubsetPercentage.setObjectName("spinBoxTrainSubsetPercentage")
        self.horizontalLayout_15.addWidget(self.spinBoxTrainSubsetPercentage)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_15.addItem(spacerItem3)
        self.verticalLayout_6.addLayout(self.horizontalLayout_15)
        self.horizontalLayout_9.addLayout(self.verticalLayout_6)
        self.tabWidget = QtWidgets.QTabWidget(self.groupBoxKShapeParams)
        self.tabWidget.setTabPosition(QtWidgets.QTabWidget.North)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.checkBoxUseGridsearch = QtWidgets.QCheckBox(self.tab)
        self.checkBoxUseGridsearch.setChecked(True)
        self.checkBoxUseGridsearch.setObjectName("checkBoxUseGridsearch")
        self.verticalLayout_5.addWidget(self.checkBoxUseGridsearch)
        self.horizontalLayout_18 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_18.setObjectName("horizontalLayout_18")
        self.label_15 = QtWidgets.QLabel(self.tab)
        self.label_15.setObjectName("label_15")
        self.horizontalLayout_18.addWidget(self.label_15)
        self.spinBoxNPartMin = QtWidgets.QSpinBox(self.tab)
        self.spinBoxNPartMin.setMinimum(2)
        self.spinBoxNPartMin.setObjectName("spinBoxNPartMin")
        self.horizontalLayout_18.addWidget(self.spinBoxNPartMin)
        self.spinBoxNPartMax = QtWidgets.QSpinBox(self.tab)
        self.spinBoxNPartMax.setMinimum(2)
        self.spinBoxNPartMax.setProperty("value", 10)
        self.spinBoxNPartMax.setObjectName("spinBoxNPartMax")
        self.horizontalLayout_18.addWidget(self.spinBoxNPartMax)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_18.addItem(spacerItem4)
        self.verticalLayout_5.addLayout(self.horizontalLayout_18)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_16 = QtWidgets.QLabel(self.tab)
        self.label_16.setObjectName("label_16")
        self.horizontalLayout_6.addWidget(self.label_16)
        self.spinBoxNCombinations = QtWidgets.QSpinBox(self.tab)
        self.spinBoxNCombinations.setMinimum(1)
        self.spinBoxNCombinations.setMaximum(10)
        self.spinBoxNCombinations.setProperty("value", 2)
        self.spinBoxNCombinations.setObjectName("spinBoxNCombinations")
        self.horizontalLayout_6.addWidget(self.spinBoxNCombinations)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem5)
        self.horizontalLayout_8.addLayout(self.horizontalLayout_6)
        self.verticalLayout_5.addLayout(self.horizontalLayout_8)
        self.horizontalLayout_20 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_20.setObjectName("horizontalLayout_20")
        self.label_17 = QtWidgets.QLabel(self.tab)
        self.label_17.setObjectName("label_17")
        self.horizontalLayout_20.addWidget(self.label_17)
        self.comboBoxSortBy = QtWidgets.QComboBox(self.tab)
        self.comboBoxSortBy.setObjectName("comboBoxSortBy")
        self.comboBoxSortBy.addItem("")
        self.comboBoxSortBy.addItem("")
        self.horizontalLayout_20.addWidget(self.comboBoxSortBy)
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_20.addItem(spacerItem6)
        self.verticalLayout_5.addLayout(self.horizontalLayout_20)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.checkBoxUseSingle = QtWidgets.QCheckBox(self.tab_2)
        self.checkBoxUseSingle.setObjectName("checkBoxUseSingle")
        self.verticalLayout_3.addWidget(self.checkBoxUseSingle)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(self.tab_2)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.spinBoxN_clusters = QtWidgets.QSpinBox(self.tab_2)
        self.spinBoxN_clusters.setMinimum(2)
        self.spinBoxN_clusters.setProperty("value", 3)
        self.spinBoxN_clusters.setObjectName("spinBoxN_clusters")
        self.horizontalLayout.addWidget(self.spinBoxN_clusters)
        spacerItem7 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem7)
        self.horizontalLayout_7.addLayout(self.horizontalLayout)
        self.verticalLayout_3.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_5 = QtWidgets.QLabel(self.tab_2)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.spinBoxRandom = QtWidgets.QSpinBox(self.tab_2)
        self.spinBoxRandom.setEnabled(False)
        self.spinBoxRandom.setMaximum(999)
        self.spinBoxRandom.setObjectName("spinBoxRandom")
        self.horizontalLayout_5.addWidget(self.spinBoxRandom)
        self.checkBoxRandom = QtWidgets.QCheckBox(self.tab_2)
        self.checkBoxRandom.setChecked(True)
        self.checkBoxRandom.setObjectName("checkBoxRandom")
        self.horizontalLayout_5.addWidget(self.checkBoxRandom)
        spacerItem8 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem8)
        self.verticalLayout_3.addLayout(self.horizontalLayout_5)
        self.tabWidget.addTab(self.tab_2, "")
        self.horizontalLayout_9.addWidget(self.tabWidget)
        self.verticalLayout_7.addLayout(self.horizontalLayout_9)
        self.verticalLayout_4.addWidget(self.groupBoxKShapeParams)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.pushButtonStart = QtWidgets.QPushButton(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.pushButtonStart.setFont(font)
        self.pushButtonStart.setObjectName("pushButtonStart")
        self.horizontalLayout_11.addWidget(self.pushButtonStart)
        self.pushButtonAbort = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonAbort.setEnabled(False)
        font = QtGui.QFont()
        font.setPointSize(14)
        font.setBold(True)
        font.setWeight(75)
        self.pushButtonAbort.setFont(font)
        self.pushButtonAbort.setObjectName("pushButtonAbort")
        self.horizontalLayout_11.addWidget(self.pushButtonAbort)
        self.verticalLayout_4.addLayout(self.horizontalLayout_11)
        self.label_6 = QtWidgets.QLabel(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_6.setFont(font)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_4.addWidget(self.label_6)
        self.textBrowser = QtWidgets.QTextBrowser(self.dockWidgetContents)
        self.textBrowser.setObjectName("textBrowser")
        self.verticalLayout_4.addWidget(self.textBrowser)
        self.groupBoxPlotParams = QtWidgets.QGroupBox(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.groupBoxPlotParams.setFont(font)
        self.groupBoxPlotParams.setObjectName("groupBoxPlotParams")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBoxPlotParams)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.label_7 = QtWidgets.QLabel(self.groupBoxPlotParams)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_14.addWidget(self.label_7)
        self.listWidgetClusterNumber = QtWidgets.QListWidget(self.groupBoxPlotParams)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.listWidgetClusterNumber.sizePolicy().hasHeightForWidth())
        self.listWidgetClusterNumber.setSizePolicy(sizePolicy)
        self.listWidgetClusterNumber.setMinimumSize(QtCore.QSize(0, 33))
        self.listWidgetClusterNumber.setMaximumSize(QtCore.QSize(16777215, 33))
        self.listWidgetClusterNumber.setProperty("showDropIndicator", False)
        self.listWidgetClusterNumber.setDefaultDropAction(QtCore.Qt.IgnoreAction)
        self.listWidgetClusterNumber.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.listWidgetClusterNumber.setFlow(QtWidgets.QListView.LeftToRight)
        self.listWidgetClusterNumber.setObjectName("listWidgetClusterNumber")
        self.horizontalLayout_14.addWidget(self.listWidgetClusterNumber)
        self.verticalLayout_2.addLayout(self.horizontalLayout_14)
        self.line = QtWidgets.QFrame(self.groupBoxPlotParams)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_2.addWidget(self.line)
        self.label_12 = QtWidgets.QLabel(self.groupBoxPlotParams)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.verticalLayout_2.addWidget(self.label_12)
        self.checkBoxShowKShapeCenters = QtWidgets.QCheckBox(self.groupBoxPlotParams)
        self.checkBoxShowKShapeCenters.setObjectName("checkBoxShowKShapeCenters")
        self.verticalLayout_2.addWidget(self.checkBoxShowKShapeCenters)
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.label_9 = QtWidgets.QLabel(self.groupBoxPlotParams)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_13.addWidget(self.label_9)
        self.spinBoxMaxNumCurves = QtWidgets.QSpinBox(self.groupBoxPlotParams)
        self.spinBoxMaxNumCurves.setMaximum(1000)
        self.spinBoxMaxNumCurves.setSingleStep(25)
        self.spinBoxMaxNumCurves.setProperty("value", 100)
        self.spinBoxMaxNumCurves.setObjectName("spinBoxMaxNumCurves")
        self.horizontalLayout_13.addWidget(self.spinBoxMaxNumCurves)
        spacerItem9 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_13.addItem(spacerItem9)
        self.verticalLayout_2.addLayout(self.horizontalLayout_13)
        self.label_13 = QtWidgets.QLabel(self.groupBoxPlotParams)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_13.setFont(font)
        self.label_13.setObjectName("label_13")
        self.verticalLayout_2.addWidget(self.label_13)
        self.horizontalLayout_16 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_16.setObjectName("horizontalLayout_16")
        self.label_10 = QtWidgets.QLabel(self.groupBoxPlotParams)
        self.label_10.setObjectName("label_10")
        self.horizontalLayout_16.addWidget(self.label_10)
        self.comboBoxErrorBand = QtWidgets.QComboBox(self.groupBoxPlotParams)
        self.comboBoxErrorBand.setObjectName("comboBoxErrorBand")
        self.comboBoxErrorBand.addItem("")
        self.comboBoxErrorBand.addItem("")
        self.comboBoxErrorBand.addItem("")
        self.horizontalLayout_16.addWidget(self.comboBoxErrorBand)
        spacerItem10 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_16.addItem(spacerItem10)
        self.verticalLayout_2.addLayout(self.horizontalLayout_16)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_14 = QtWidgets.QLabel(self.groupBoxPlotParams)
        self.label_14.setObjectName("label_14")
        self.verticalLayout.addWidget(self.label_14)
        self.horizontalLayout_17 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_17.setObjectName("horizontalLayout_17")
        self.radioButtonXZeroZero = QtWidgets.QRadioButton(self.groupBoxPlotParams)
        self.radioButtonXZeroZero.setChecked(True)
        self.radioButtonXZeroZero.setObjectName("radioButtonXZeroZero")
        self.horizontalLayout_17.addWidget(self.radioButtonXZeroZero)
        self.radioButtonXZeroMaxima = QtWidgets.QRadioButton(self.groupBoxPlotParams)
        self.radioButtonXZeroMaxima.setObjectName("radioButtonXZeroMaxima")
        self.horizontalLayout_17.addWidget(self.radioButtonXZeroMaxima)
        spacerItem11 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_17.addItem(spacerItem11)
        self.verticalLayout.addLayout(self.horizontalLayout_17)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.pushButtonApplyPlotOptions = QtWidgets.QPushButton(self.groupBoxPlotParams)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.pushButtonApplyPlotOptions.setFont(font)
        self.pushButtonApplyPlotOptions.setObjectName("pushButtonApplyPlotOptions")
        self.verticalLayout_2.addWidget(self.pushButtonApplyPlotOptions)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.pushButtonSave = QtWidgets.QPushButton(self.groupBoxPlotParams)
        self.pushButtonSave.setObjectName("pushButtonSave")
        self.gridLayout.addWidget(self.pushButtonSave, 0, 0, 1, 1)
        self.pushButtonLoad = QtWidgets.QPushButton(self.groupBoxPlotParams)
        self.pushButtonLoad.setObjectName("pushButtonLoad")
        self.gridLayout.addWidget(self.pushButtonLoad, 0, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.verticalLayout_4.addWidget(self.groupBoxPlotParams)
        KShapeControl.setWidget(self.dockWidgetContents)

        self.retranslateUi(KShapeControl)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(KShapeControl)
        KShapeControl.setTabOrder(self.pushButtonReconnectFlowchartInput, self.comboBoxDataColumn)
        KShapeControl.setTabOrder(self.comboBoxDataColumn, self.spinBoxMaxIter)
        KShapeControl.setTabOrder(self.spinBoxMaxIter, self.spinBoxTol)
        KShapeControl.setTabOrder(self.spinBoxTol, self.spinBoxTrainSubsetPercentage)
        KShapeControl.setTabOrder(self.spinBoxTrainSubsetPercentage, self.tabWidget)
        KShapeControl.setTabOrder(self.tabWidget, self.checkBoxUseGridsearch)
        KShapeControl.setTabOrder(self.checkBoxUseGridsearch, self.spinBoxNPartMin)
        KShapeControl.setTabOrder(self.spinBoxNPartMin, self.spinBoxNPartMax)
        KShapeControl.setTabOrder(self.spinBoxNPartMax, self.spinBoxNCombinations)
        KShapeControl.setTabOrder(self.spinBoxNCombinations, self.comboBoxSortBy)
        KShapeControl.setTabOrder(self.comboBoxSortBy, self.pushButtonStart)
        KShapeControl.setTabOrder(self.pushButtonStart, self.pushButtonAbort)
        KShapeControl.setTabOrder(self.pushButtonAbort, self.textBrowser)
        KShapeControl.setTabOrder(self.textBrowser, self.listWidgetClusterNumber)
        KShapeControl.setTabOrder(self.listWidgetClusterNumber, self.checkBoxShowKShapeCenters)
        KShapeControl.setTabOrder(self.checkBoxShowKShapeCenters, self.spinBoxMaxNumCurves)
        KShapeControl.setTabOrder(self.spinBoxMaxNumCurves, self.comboBoxErrorBand)
        KShapeControl.setTabOrder(self.comboBoxErrorBand, self.radioButtonXZeroZero)
        KShapeControl.setTabOrder(self.radioButtonXZeroZero, self.radioButtonXZeroMaxima)
        KShapeControl.setTabOrder(self.radioButtonXZeroMaxima, self.pushButtonApplyPlotOptions)
        KShapeControl.setTabOrder(self.pushButtonApplyPlotOptions, self.pushButtonSave)
        KShapeControl.setTabOrder(self.pushButtonSave, self.pushButtonLoad)
        KShapeControl.setTabOrder(self.pushButtonLoad, self.spinBoxN_init)
        KShapeControl.setTabOrder(self.spinBoxN_init, self.spinBoxN_clusters)
        KShapeControl.setTabOrder(self.spinBoxN_clusters, self.checkBoxUseSingle)
        KShapeControl.setTabOrder(self.checkBoxUseSingle, self.checkBoxRandom)
        KShapeControl.setTabOrder(self.checkBoxRandom, self.spinBoxRandom)

    def retranslateUi(self, KShapeControl):
        _translate = QtCore.QCoreApplication.translate
        KShapeControl.setWindowTitle(_translate("KShapeControl", "Co&ntrols"))
        self.groupBoxKShapeParams.setTitle(_translate("KShapeControl", "KShape params"))
        self.pushButtonReconnectFlowchartInput.setText(_translate("KShapeControl", "Reconnect flowchart input"))
        self.labelConnectedToFlowchart.setText(_translate("KShapeControl", "INPUT CONNECTED TO FLOWCHART"))
        self.label_8.setText(_translate("KShapeControl", "data_column"))
        self.label_2.setText(_translate("KShapeControl", "max_iter:"))
        self.label_3.setText(_translate("KShapeControl", "tol: 10 ^"))
        self.label_4.setText(_translate("KShapeControl", "n_init:"))
        self.label_11.setText(_translate("KShapeControl", "training subset (%)"))
        self.checkBoxUseGridsearch.setText(_translate("KShapeControl", "Use Gridsearch"))
        self.label_15.setToolTip(_translate("KShapeControl", "Range of partitions (and therefore n_clusters) to try"))
        self.label_15.setText(_translate("KShapeControl", "npart rng:"))
        self.label_16.setToolTip(_translate("KShapeControl", "Number of seed combinations to try out"))
        self.label_16.setText(_translate("KShapeControl", "ncombs: 10^"))
        self.label_17.setToolTip(_translate("KShapeControl", "Sort the input arrays by half-peak-width or amplitude before partitioning to pick centroid seeds"))
        self.label_17.setText(_translate("KShapeControl", "sortby"))
        self.comboBoxSortBy.setItemText(0, _translate("KShapeControl", "width"))
        self.comboBoxSortBy.setItemText(1, _translate("KShapeControl", "amplitude"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("KShapeControl", "Grid"))
        self.checkBoxUseSingle.setText(_translate("KShapeControl", "Use Single"))
        self.label.setText(_translate("KShapeControl", "n_clusters:"))
        self.label_5.setText(_translate("KShapeControl", "rnd_state:"))
        self.checkBoxRandom.setText(_translate("KShapeControl", "random"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("KShapeControl", "Single"))
        self.pushButtonStart.setText(_translate("KShapeControl", "Start"))
        self.pushButtonAbort.setText(_translate("KShapeControl", "Abort"))
        self.label_6.setText(_translate("KShapeControl", "Status:"))
        self.textBrowser.setHtml(_translate("KShapeControl", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Noto Sans\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.groupBoxPlotParams.setTitle(_translate("KShapeControl", "Plot Options"))
        self.label_7.setText(_translate("KShapeControl", "Plot cluster:"))
        self.label_12.setText(_translate("KShapeControl", "Raw plots"))
        self.checkBoxShowKShapeCenters.setText(_translate("KShapeControl", "Show centers"))
        self.label_9.setText(_translate("KShapeControl", "max num curves:"))
        self.label_13.setText(_translate("KShapeControl", "Mean plots"))
        self.label_10.setText(_translate("KShapeControl", "Error band"))
        self.comboBoxErrorBand.setItemText(0, _translate("KShapeControl", "standard deviation"))
        self.comboBoxErrorBand.setItemText(1, _translate("KShapeControl", "confidence interval"))
        self.comboBoxErrorBand.setItemText(2, _translate("KShapeControl", "None"))
        self.label_14.setText(_translate("KShapeControl", "Set x = 0 at:"))
        self.radioButtonXZeroZero.setText(_translate("KShapeControl", "&Zero"))
        self.radioButtonXZeroMaxima.setText(_translate("KShapeControl", "&Maxima"))
        self.pushButtonApplyPlotOptions.setText(_translate("KShapeControl", "Apply Options"))
        self.pushButtonSave.setText(_translate("KShapeControl", "Save"))
        self.pushButtonLoad.setText(_translate("KShapeControl", "Load"))

